from pydantic import BaseModel, Field
from typing import List, Dict, Literal, Union, Optional, Tuple


class MultipoleGeoElement(BaseModel):
    """
    Level 5: Class for FiQuS Multipole
    """
    lines: Optional[int] = Field(
        default=3,
        description="It specifies the number of Gaussian points for lines.",
    )
    triangles: Optional[Literal[1, 3, 4, 6, 7, 12, 13, 16]] = Field(
        default=3,
        description="It specifies the number of Gaussian points for triangles.",
    )
    quadrangles: Optional[Literal[1, 3, 4, 7]] = Field(
        default=4,
        description="It specifies the number of Gaussian points for quadrangles.",
    )


class MultipoleSolveConvectionBoundaryCondition(BaseModel):
    """
    Level 5: Class for FiQuS Multipole
    """
    boundaries: Optional[List[str]] = Field(
        default=[],
        description="It specifies the list of boundaries where the condition is applied."
                    "Each boundary is identified by a string of the form <half-turn/wedge reference number><side>,"
                    "where the accepted sides are i, o, l, h which correspond respectively to inner, outer, lower (angle), higher (angle): e.g., 1o",
    )
    heat_transfer_coefficient: Optional[Union[float, str]] = Field(
        default=None,
        description="It specifies the value or function name of the heat transfer coefficient for this boundary condition.",
    )


class MultipoleSolveHeatFluxBoundaryCondition(BaseModel):
    """
    Level 5: Class for FiQuS Multipole
    """
    boundaries: Optional[List[str]] = Field(
        default=[],
        description="It specifies the list of boundaries where the condition is applied."
                    "Each boundary is identified by a string of the form <half-turn/wedge reference number><side>,"
                    "where the accepted sides are i, o, l, h which correspond respectively to inner, outer, lower (angle), higher (angle): e.g., 1o",
    )
    const_heat_flux: Optional[float] = Field(
        default=None,
        description="It specifies the value of the heat flux for this boundary condition.",
    )
    # function_heat_flux: Optional[str] = None


class MultipoleSolveTemperatureBoundaryCondition(BaseModel):
    """
    Level 5: Class for FiQuS Multipole
    """
    boundaries: Optional[List[str]] = Field(
        default=[],
        description="It specifies the list of boundaries where the condition is applied."
                    "Each boundary is identified by a string of the form <half-turn/wedge reference number><side>,"
                    "where the accepted sides are i, o, l, h which correspond respectively to inner, outer, lower (angle), higher (angle): e.g., 1o",
    )
    const_temperature: Optional[float] = Field(
        default=None,
        description="It specifies the value of the temperature for this boundary condition.",
    )
    # function_temperature: Optional[str] = None


class MultipoleSolveQuenchInitiation(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    """
    turns: Optional[List[int]] = Field(
        default=[],
        description="It specifies the list of reference numbers of half-turns whose critical currents are set to zero.",
    )
    t_trigger: Optional[List[float]] = Field(
        default=[],
        description="It specifies the list of time instants at which the critical current is set to zero.",
    )


class MultipoleSolveBoundaryConditionsThermal(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    """
    temperature: Optional[Dict[str, MultipoleSolveTemperatureBoundaryCondition]] = Field(
        default={},
        description="This dictionary contains the information about the Dirichlet boundary conditions."
                    "The keys are chosen names for each boundary condition.",
    )
    heat_flux: Optional[Dict[str, MultipoleSolveHeatFluxBoundaryCondition]] = Field(
        default={},
        description="This dictionary contains the information about the Neumann boundary conditions."
                    "The keys are chosen names for each boundary condition.",
    )
    cooling: Optional[Dict[str, MultipoleSolveConvectionBoundaryCondition]] = Field(
        default={},
        description="This dictionary contains the information about the Robin boundary conditions."
                    "The keys are chosen names for each boundary condition.",
    )


# class MultipoleSolveTransientElectromagnetics(BaseModel):
#     """
#     Level 4: Class for FiQuS Multipole
#     """
#     time_stepping: Optional[Literal["adaptive", "fixed"]] = Field(
#         default="adaptive",
#         description="It specifies the type of time stepping.",
#     )
#     initial_time: Optional[float] = Field(
#         default=0.,
#         description="It specifies the initial time of the simulation.",
#     )
#     final_time: Optional[float] = Field(
#         default=None,
#         description="It specifies the final time of the simulation.",
#     )
#     fixed: MultipoleSolveTimeParametersFixed = Field(
#         default=MultipoleSolveTimeParametersFixed(),
#         description="This dictionary contains the information about the time parameters of the fixed time stepping.",
#     )
#     adaptive: MultipoleSolveTimeParametersAdaptive = Field(
#         default=MultipoleSolveTimeParametersAdaptive(),
#         description="This dictionary contains the information about the time parameters of the adaptive time stepping.",
#     )


class MultipoleSolveHeCooling(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    """
    enabled: Optional[bool] = Field(
        default=False,
        description="It determines whether the helium cooling is enabled or not (adiabatic conditions).",
    )
    sides: Optional[Literal["external", "inner", "outer", "inner_outer"]] = Field(
        default="outer",
        description="It specifies the general grouping of the boundaries where to apply cooling:"
                    "'external': all external boundaries; 'inner': only inner boundaries; 'outer': only outer boundaries; 'inner_outer': inner and outer boundaries.",
    )
    heat_transfer_coefficient: Optional[Union[float, str]] = Field(
        default=0.0,
        description="It specifies the value or name of the function of the constant heat transfer coefficient.",
    )

class MultipoleSolveNonLinearSolver(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    """
    rel_tolerance: Optional[float] = Field(
        default=1E-4,
        description="It specifies the relative tolerance.",
    )
    abs_tolerance: Optional[float] = Field(
        default=0.1,
        description="It specifies the absolute tolerance.",
    )
    relaxation_factor: Optional[float] = Field(
        default=0.7,
        description="It specifies the relaxation factor.",
    )
    max_iterations: Optional[int] = Field(
        default=20,
        description="It specifies the maximum number of iterations if no convergence is reached.",
    )
    norm_type: Literal["L1Norm", "MeanL1Norm", "L2Norm", "MeanL2Norm", "LinfNorm"] = Field(
        default='LinfNorm',
        description="It specifies the type of norm to be calculated for convergence assessment.",
    )


class MultipoleSolveTransientThermal(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    """
    initial_time: Optional[float] = Field(
        default=0.,
        description="It specifies the initial time of the simulation.",
    )
    final_time: Optional[float] = Field(
        default=None,
        description="It specifies the final time of the simulation.",
    )
    initial_time_step: Optional[float] = Field(
        default=1E-10,
        description="It specifies the initial time step used at the beginning of the transient simulation.",
    )
    min_time_step: Optional[float] = Field(
        default=1E-12,
        description="It specifies the minimum possible value of the time step.",
    )
    max_time_step: Optional[float] = Field(
        default=10,
        description="It specifies the maximum possible value of the time step.",
    )
    breakpoints: Optional[List[float]] = Field(
        default=[],
        description="It forces the transient simulation to hit the time instants contained in this list.",
    )
    integration_method: Union[None, Literal[
       "Euler", "Gear_2", "Gear_3", "Gear_4", "Gear_5", "Gear_6"
    ]] = Field(
        default="Euler",
        title="Integration Method",
        description="It specifies the type of integration method to be used.",
    )
    rel_tol_time: Optional[float] = Field(
        default=1E-4,
        description="It specifies the relative tolerance.",
    )
    abs_tol_time: Optional[float] = Field(
        default=1e-4,
        description="It specifies the absolute tolerance.",
    )
    norm_type: Literal["L1Norm", "MeanL1Norm", "L2Norm", "MeanL2Norm", "LinfNorm"] = Field(
        default='LinfNorm',
        description="It specifies the type of norm to be calculated for convergence assessment.",
    )
    stop_temperature: Optional[float] = Field(
        default=300,
        description="If one half turn reaches this temperature, the simulation is stopped.",
    )


class MultipoleSolveInsulationBlockToBlock(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    It contains the information about the materials and thicknesses of the inner insulation regions (between blocks) modeled via thin-shell approximation.
    """
    material: Optional[str] = Field(
        default=None,
        description="It specifies the default material of the insulation regions between the blocks insulation regions.",
    )
    # the order of blocks should be: [inner, outer] for mid-layer couples or [lower, higher] for mid-pole and mid-winding couples
    blocks_connection_overwrite: List[Tuple[str, str]] = Field(
        default=[],
        description="It specifies the blocks couples adjacent to the insulation region."
                    "The blocks must be ordered from inner to outer block for mid-layer insulation regions and from lower to higher angle block for mid-pole and mid-winding insulation regions.",
    )
    materials_overwrite: Optional[List[List[str]]] = Field(
        default=[],
        description="It specifies the list of materials making up the layered insulation region to be placed between the specified blocks."
                    "The materials must be ordered from inner to outer layers and lower to higher angle layers.",
    )
    thicknesses_overwrite: Optional[List[List[Optional[float]]]] = Field(
        default=[],
        description="It specifies the list of thicknesses of the specified insulation layers. The order must match the one of the materials list.",
    )


class MultipoleSolveInsulationExterior(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    It contains the information about the materials and thicknesses of the outer insulation regions (exterior boundaries) modeled via thin-shell approximation.
    """
    blocks: Optional[List[str]] = Field(
        default=[],
        description="It specifies the reference numbers of the blocks adjacent to the exterior insulation regions to modify.",
    )
    materials_append: Optional[List[List[str]]] = Field(
        default=[],
        description="It specifies the list of materials making up the layered insulation region to be appended to the block insulation."
                    "The materials must be ordered from the block outward.",
    )
    thicknesses_append: Optional[List[List[float]]] = Field(
        default=[],
        description="It specifies the list of thicknesses of the specified insulation layers. The order must match the one of the materials list.",
    )


class MultipoleSolveWedge(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    material: Optional[str] = Field(
        default=None,
        description="It specifies the material of the wedge regions.",
    )
    RRR: Optional[float] = Field(
        default=None,
        description="It specifies the RRR of the wedge regions.",
    )
    T_ref_RRR_high: Optional[float] = Field(
        default=None,
        description="It specifies the reference temperature associated with the RRR.",
    )


class MultipoleSolveInsulationTSA(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    block_to_block: MultipoleSolveInsulationBlockToBlock = Field(
        default=MultipoleSolveInsulationBlockToBlock(),
        description="This dictionary contains the information about the materials and thicknesses of the inner insulation regions (between blocks) modeled via thin-shell approximation.",
    )
    exterior: Optional[MultipoleSolveInsulationExterior] = Field(
        default=MultipoleSolveInsulationExterior(),
        description="This dictionary contains the information about the materials and thicknesses of the outer insulation regions (exterior boundaries) modeled via thin-shell approximation.",
    )


class MultipoleSolveThermal(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    solve_type: Optional[Literal[None, "transient"]] = Field(
        default=None,
        description="It determines whether the thermal transient problem is solved ('transient') or not ('null').",
    )
    insulation_TSA: Optional[MultipoleSolveInsulationTSA] = Field(
        default=MultipoleSolveInsulationTSA(),
        description="This dictionary contains the information about the materials and thicknesses of the insulation regions modeled via thin-shell approximation.",
    )
    He_cooling: MultipoleSolveHeCooling = Field(
        default=MultipoleSolveHeCooling(),
        description="This dictionary contains the information about the Robin boundary condition for generic groups of boundaries.",
    )
    overwrite_boundary_conditions: Optional[MultipoleSolveBoundaryConditionsThermal] = Field(
        default=MultipoleSolveBoundaryConditionsThermal(),
        description="This dictionary contains the information about boundary conditions for explicitly specified boundaries.",
    )
    non_linear_solver: MultipoleSolveNonLinearSolver = Field(
        default=MultipoleSolveNonLinearSolver(),
        description="This dictionary contains the information about the parameters for the non-linear solver.",
    )
    time_stepping: MultipoleSolveTransientThermal = Field(
        default=MultipoleSolveTransientThermal(),
        description="This dictionary contains the information about the parameters for the transient solver.",
    )
    jc_degradation_to_zero: Optional[MultipoleSolveQuenchInitiation] = Field(
        default=MultipoleSolveQuenchInitiation(),
        description="This dictionary contains the information about half turns with zero critical current.",
    )
    init_temperature: Optional[float] = Field(
        default=1.9,
        description="It specifies the initial temperature of the simulation.",
    )
    enforce_init_temperature_as_minimum: Optional[bool] = Field(
        default=False,
        description="It determines whether the initial temperature is enforced as the minimum temperature of the simulation.",
    )

class MultipoleSolveElectromagnetics(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    solve_type: Optional[Literal[None, "stationary"]] = Field(
        default=None,
        description="It determines whether the magneto-static problem is solved ('stationary') or not ('null').",
    )

    non_linear_solver: MultipoleSolveNonLinearSolver = Field(
        default=MultipoleSolveNonLinearSolver(),
        description="This dictionary contains the information about the parameters for the non-linear solver.",
    )
    # currently not needed since stationary only, we will be able to reuse it from the thermal solver
    # time_stepping_parameters: MultipoleSolveTransientElectromagnetics = Field(
    #     default=MultipoleSolveTransientElectromagnetics(),
    #     description="This dictionary contains the information about the parameters for the transient solver.",
    # )


class MultipoleMeshThinShellApproximationParameters(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    """
    minimum_discretizations: Optional[int] = Field(
        default=1,
        description="It specifies the number of minimum spacial discretizations across a thin-shell.",
    )
    global_size_QH: Optional[float] = Field(
        default=1e-4,
        description="The thickness of the quench heater region is divided by this parameter to determine the number of spacial discretizations across the thin-shell.",
    )
    minimum_discretizations_QH: Optional[int] = Field(
        default=1,
        description="It specifies the number of minimum spacial discretizations across a thin-shell.",
    )


class MultipoleMeshThreshold(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    enabled: Optional[bool] = Field(
        default=False,
        description="It determines whether the gmsh Field is enabled or not.",
    )
    SizeMin: Optional[float] = Field(
        default=None,
        description="It sets gmsh Mesh.MeshSizeMin.",
    )
    SizeMax: Optional[float] = Field(
        default=None,
        description="It sets gmsh Mesh.MeshSizeMax.",
    )
    DistMin: Optional[float] = Field(
        default=None,
        description="It sets gmsh Mesh.MeshDistMin.",
    )
    DistMax: Optional[float] = Field(
        default=None,
        description="It sets gmsh Mesh.MeshDistMax.",
    )

class MultipoleMeshTransfinite(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    enabled_for: Literal[None, "curves", "curves_and_surfaces"] = Field(
        default=None,
        description="It determines on what entities the transfinite algorithm is applied.",
    )
    curve_target_size_height: Optional[float] = Field(
        default=1.0,
        description="The height of the region (short side) is divided by this parameter to determine the number of elements to apply via transfinite curves.",
    )
    curve_target_size_width: Optional[float] = Field(
        default=1.0,
        description="The width of the region (long side) is divided by this parameter to determine the number of elements to apply via transfinite curves.",
    )
class MultipoleMeshTransfiniteOrField(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    transfinite: MultipoleMeshTransfinite = Field(
        default=MultipoleMeshTransfinite(),
        description="This dictionary contains the mesh information for transfinite curves.",
    )
    field: MultipoleMeshThreshold = Field(
        default=MultipoleMeshThreshold(),
        description="This dictionary contains the gmsh Field information.",
    )

class MultipolePostProcThermal(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    output_time_steps_pos: Optional[Union[bool, int]] = Field(
        default=True,
        description="It determines whether the solution for the .pos file is saved for all time steps (True), none (False), or equidistant time steps (int).",
    )
    output_time_steps_txt: Optional[Union[bool, int]] = Field(
        default=True,
        description="It determines whether the solution for the .txt file is saved for all time steps (True), none (False), or equidistant time steps (int).",
    )
    save_pos_at_the_end: Optional[bool] = Field(
        default=True,
        description="It determines whether the solution for the .pos file is saved at the end of the simulation or during run time.",
    )
    save_txt_at_the_end: Optional[bool] = Field(
        default=False,
        description="It determines whether the solution for the .txt file is saved at the end of the simulation or during run time.",
    )
    take_average_conductor_temperature: Optional[bool] = Field(
        default=True,
        description="It determines whether the output files are based on the average conductor temperature or not (map2d).",
    )
    plot_all: Optional[Union[bool, None]] = Field(
        default=False,
        description="It determines whether the figures are generated and shown (true), generated only (null), or not generated (false). Useful for tests.",
    )
    variables: Optional[List[Literal["T", "jOverJc", "rho"]]] = Field(
        default=["T"],
        description="It specifies the physical quantity to be output.",
    )
    volumes: Optional[List[
        Literal["omega", "powered", "induced", "iron", "conducting", "insulator"]]] = Field(
        default=["powered"],
        description="It specifies the regions associated with the physical quantity to be output.",
    )


class MultipolePostProcElectromagnetics(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    output_time_steps_pos: Optional[Union[bool, int]] = Field(
        default=True,
        description="It determines whether the solution for the .pos file is saved for all time steps (True), none (False), or equidistant time steps (int).",
    )
    output_time_steps_txt: Optional[Union[bool, int]] = Field(
        default=True,
        description="It determines whether the solution for the .txt file is saved for all time steps (True), none (False), or equidistant time steps (int).",
    )
    save_pos_at_the_end: Optional[bool] = Field(
        default=True,
        description="It determines whether the solution for the .pos file is saved at the end of the simulation or during run time.",
    )
    save_txt_at_the_end: Optional[bool] = Field(
        default=False,
        description="It determines whether the solution for the .txt file is saved at the end of the simulation or during run time.",
    )
    compare_to_ROXIE: Optional[str] = Field(
        default=None,
        description="It contains the absolute path to a reference ROXIE map2d file. If provided, comparative plots with respect to the reference are generated.",
    )
    plot_all: Optional[Union[bool, None]] = Field(
        default=False,
        description="It determines whether the figures are generated and shown (true), generated only (null), or not generated (false). Useful for tests.",
    )
    variables: Optional[List[Literal["a", "az", "b", "h", "js"]]] = Field(
        default=["b"],
        description="It specifies the physical quantity to be output.",
    )
    volumes: Optional[List[
        Literal["omega", "powered", "induced", "air", "air_far_field", "iron", "conducting", "insulator"]]] = Field(
        default=["powered"],
        description="It specifies the regions associated with the physical quantity to be output.",
    )


class MultipolePostProc(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    electromagnetics: MultipolePostProcElectromagnetics = Field(
        default=MultipolePostProcElectromagnetics(),
        description="This dictionary contains the post-processing information for the electromagnetic solution.",
    )
    thermal: MultipolePostProcThermal = Field(
        default=MultipolePostProcThermal(),
        description="This dictionary contains the post-processing information for the thermal solution.",
    )


class MultipoleSolve(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    electromagnetics: MultipoleSolveElectromagnetics = Field(
        default=MultipoleSolveElectromagnetics(),
        description="This dictionary contains the solver information for the electromagnetic solution.",
    )
    thermal: MultipoleSolveThermal = Field(
        default=MultipoleSolveThermal(),
        description="This dictionary contains the solver information for the thermal solution.",
    )
    wedges: MultipoleSolveWedge = Field(
        default=MultipoleSolveWedge(),
        description="This dictionary contains the material information of wedges.",
    )
    noOfMPITasks: Optional[Union[bool, int]] = Field(
        default=False,
        title="No. of tasks for MPI parallel run of GetDP",
        description=(
            "If integer, GetDP will be run in parallel using MPI. This is only valid"
            " if MPI is installed on the system and an MPI-enabled GetDP is used." 
            " If False, GetDP will be run in serial without invoking mpiexec."
        ),
    )

class MultipoleThermalInsulationMesh(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    global_size: float = Field(
        default=1e-4,
        description="It specifies the global size of the mesh for the insulation regions. It is enforced as a constant mesh field for surface insulation and by fixing the number of TSA layers for thin-shell approximation.",
    )
    TSA: Optional[MultipoleMeshThinShellApproximationParameters] = Field(
        default=MultipoleMeshThinShellApproximationParameters(),
        description="This dictionary contains the mesh information for thin-shells.",
    )
    
class MultipoleMeshThermal(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    create: bool = Field(
        default=True,
        description="It determines whether the thermal mesh is built or not.",
    )
    conductors: Optional[MultipoleMeshTransfiniteOrField] = Field(
        default=MultipoleMeshTransfiniteOrField(),
        description="This dictionary contains the mesh information for the conductor regions.",
    )
    wedges: Optional[MultipoleMeshTransfiniteOrField] = Field(
        default=MultipoleMeshTransfiniteOrField(),
        description="This dictionary contains the mesh information for the wedge regions.",
    )
    iron_field: Optional[MultipoleMeshThreshold] = Field(
        default=MultipoleMeshThreshold(),
        description="This dictionary contains the gmsh Field information for the iron yoke region.",
    )
    insulation: Optional[MultipoleThermalInsulationMesh] = Field(
        default=MultipoleThermalInsulationMesh(),
        description="This dictionary contains the mesh information for the insulation regions.",
    )
   
    iron_field: Optional[MultipoleMeshThreshold] = Field(
        default=MultipoleMeshThreshold(),
        description="This dictionary contains the gmsh Field information for the iron yoke region.",
    )

    isothermal_conductors: Optional[bool] = Field(
        default=False,
        description="It determines whether the conductors are considered isothermal or not using getDP constraints.",
    )
    isothermal_wedges: Optional[bool] = Field(
        default=False,
        description="It determines whether the wedges are considered isothermal or not using getDP Link constraints.",
    )

class MultipoleMeshElectromagnetics(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    create: bool = Field(
        default=True,
        description="It determines whether the electromagnetic mesh is built or not.",
    )
    conductors: Optional[MultipoleMeshTransfiniteOrField] = Field(
        default=MultipoleMeshTransfiniteOrField(),
        description="This dictionary contains the mesh information for the conductor regions.",
    )
    wedges: Optional[MultipoleMeshTransfiniteOrField] = Field(
        default=MultipoleMeshTransfiniteOrField(),
        description="This dictionary contains the mesh information for the wedge regions.",
    )
    iron_field: Optional[MultipoleMeshThreshold] = Field(
        default=MultipoleMeshThreshold(),
        description="This dictionary contains the gmsh Field information for the iron yoke region.",
    )
    bore_field: Optional[MultipoleMeshThreshold] = Field(
        default=MultipoleMeshThreshold(),
        description="This dictionary contains the gmsh Field information for the bore region.",
    )

class MultipoleMesh(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    electromagnetics: MultipoleMeshElectromagnetics = Field(
        default=MultipoleMeshElectromagnetics(),
        description="This dictionary contains the mesh information for the electromagnetic solution.",
    )
    thermal: MultipoleMeshThermal = Field(
        default=MultipoleMeshThermal(),
        description="This dictionary contains the mesh information for the thermal solution.",
    )


class MultipoleGeometryThermal(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    create: bool = Field(
        default=True,
        description="It determines whether the thermal geometry is built or not.",
    )
    with_iron_yoke: Optional[bool] = Field(
        default=False,
        description="It determines whether the iron yoke region is built or not.",
    )
    with_wedges: Optional[bool] = Field(
        default=True,
        description="It determines whether the wedge regions are built or not.",
    )
    use_TSA: Optional[bool] = Field(
        default=False,
        description="It determines whether the insulation regions are explicitly built or modeled via thin-shell approximation.",
    )
    correct_block_coil_tsa_checkered_scheme: Optional[bool] = Field(
        default=False,
        description="There is a bug in the TSA naming scheme for block coils, this flag activates a simple (not clean) bug fix that will be replaced in a future version.",
    )

class MultipoleGeometryElectromagnetics(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    create: bool = Field(
        default=True,
        description="It determines whether the electromagnetic geometry is built or not.",
    )
    with_iron_yoke: Optional[bool] = Field(
        default=True,
        description="It determines whether the iron yoke region is built or not.",
    )
    with_wedges: Optional[bool] = Field(
        default=True,
        description="It determines whether the wedge regions are built or not.",
    )
    symmetry: Optional[Literal["none", "xy", "x", "y"]] = Field(
        default='none',
        description="It determines the model regions to build according to the specified axis/axes.",
    )


class MultipoleGeometry(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    geom_file_path: Optional[str] = Field(
        default=None,
        description="It contains the path to a .geom file. If null, the default .geom file produced by steam-sdk BuilderFiQuS will be used.",
    )
    plot_preview: Optional[bool] = Field(
        default=False,
        description="If true, it displays matplotlib figures of the magnet geometry with relevant information (e.g., conductor and block numbers).",
    )
    electromagnetics: MultipoleGeometryElectromagnetics = Field(
        default=MultipoleGeometryElectromagnetics(),
        description="This dictionary contains the geometry information for the electromagnetic solution.",
    )
    thermal: MultipoleGeometryThermal = Field(
        default=MultipoleGeometryThermal(),
        description="This dictionary contains the geometry information for the thermal solution.",
    )


class Multipole(BaseModel):
    """
    Level 1: Class for FiQuS Multipole
    """
    type: Literal["multipole"] = "multipole"
    geometry: MultipoleGeometry = Field(
        default=MultipoleGeometry(),
        description="This dictionary contains the geometry information.",
    )
    mesh: MultipoleMesh = Field(
        default=MultipoleMesh(),
        description="This dictionary contains the mesh information.",
    )
    solve: MultipoleSolve = Field(
        default=MultipoleSolve(),
        description="This dictionary contains the solution information.",
    )
    postproc: MultipolePostProc = Field(
        default=MultipolePostProc(),
        description="This dictionary contains the post-process information.",
    )



import os
import numpy as np
import re
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter
from matplotlib.animation import FuncAnimation
import pandas as pd
from scipy import integrate, interpolate
from ruamel.yaml import YAML
import subprocess

# from fiqus.utils.Utils import FilesAndFolders as Util
# from fiqus.data.DataFiQuSConductorAC_Strand import CACStrandSolve, CACStrandPostproc, CACStrandMesh, CACStrandGeometry

def create_non_overwriting_filepath(folder_path, base_name, extension, overwrite):
        """
            Creates a filepath that does not overwrite any existing files.

            This function checks if a file already exists at the specified filepath. If the file exists and `overwrite` is False, 
            it modifies the filepath to create a new file instead of overwriting the existing one. 
            If `overwrite` is True or the file does not exist, it returns the filepath as it is.

            Parameters
            ----------
            folder_path : str
                The path to the folder where the file will be created.
            base_name : str
                The base name of the file.
            extension : str
                The extension of the file.
            overwrite : bool, optional
                If True, the function will overwrite an existing file. If False, the function will modify the filepath to avoid overwriting. Defaults to False.

            Returns
            -------
            str
                The final filepath. If `overwrite` is False and a file already exists at the original filepath, this will be a new filepath that does not overwrite any existing files.
        """
        if os.path.exists(os.path.join(folder_path, base_name+extension)) and not overwrite:
            counter = 1
            new_name = base_name + f"_{counter}" + extension
            while os.path.exists(os.path.join(folder_path, new_name)):
                new_name = base_name + f"_{counter}" + extension
                counter += 1
            return os.path.join(folder_path, new_name)

        return os.path.join(folder_path, base_name+extension)

class YamlWrapper:
    """
    A wrapper class for YAML data that allows accessing dictionary data using dot notation.
    """
    def __init__(self, data):
        for key, value in data.items():
            if isinstance(value, dict):
                value = YamlWrapper(value)
            self.__dict__[key] = value

    def __getattr__(self, item):
        return self.__dict__.get(item)

    def __setattr__(self, key, value):
        self.__dict__[key] = value

def load_yaml(file_path):
    """
    Load a YAML file and return the data as a YamlWrapper object. This enables accessing the data using dot notation (e.g., data.key.subkey), without needing a predefined pydantic model, allowing for better backwards compatibility.
    """
    yaml = YAML()
    with open(file_path, 'r') as file:
        data = yaml.load(file)
    return YamlWrapper(data)

def is_latex_installed():
    """
    Check if LaTeX is installed on the system.
    """
    try:
        subprocess.run(['pdflatex', '--version'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        return True
    except subprocess.CalledProcessError:
        return False
    except FileNotFoundError:
        return False

class SimulationData:
    """
    Class used to store and manage data from a single simulation.

    This class is responsible for loading and organizing the data from a single simulation. 
    It stores the data in various attributes and provides methods for retrieving and processing the data.

    """
    def __init__(self, model_data_output_path, geometry_name, mesh_name, solution_name) -> None:
        self.model_data_output_path = model_data_output_path # This is the path to the folder where the model output data is stored (e.g. geometries)
        self.geometry_name = geometry_name # Name of the geometry folder
        self.mesh_name = mesh_name # Name of the mesh folder 
        self.solution_name = solution_name # Name of the solution folder

        # Organize the folders:
        self.geometry_folder = os.path.join(self.model_data_output_path, geometry_name) # Path to the geometry folder
        self.mesh_folder = os.path.join(self.geometry_folder, mesh_name) # Path to the mesh folder
        self.solution_folder = os.path.join(self.mesh_folder, solution_name) # Path to the solution folder

        # Store the YAML input-files in a data model, fdm:
        self.geometry, self.mesh, self.solve = self.retrieve_fiqusDataModel()

        # Store losses, simulation time and check if the simulation crashed:
        temp_file_path = os.path.join(self.solution_folder, 'test_temporary')
        loss_file = [f for f in os.listdir(temp_file_path) if f.startswith('power') and f.endswith('.txt')][0]
        self.power_columns = ['Time', 'FilamentLoss', 'CouplingLoss', 'EddyLoss', 'TotalLoss', 'CouplingLoss_dyn', 'TotalLoss_dyn'] # Only in the case dynamic correction is used, must be changed later
        self.power = pd.read_csv(os.path.join(self.solution_folder, 'test_temporary', loss_file), sep = ' ', names=self.power_columns) # Store instantaneous losses as pandas dataframe
        self.crash = True if 'crash_report.txt' in os.listdir(temp_file_path) else False

        # Add a row of zeros at the beginning of the dataframe to account for the initial condition:
        self.power = pd.concat([pd.DataFrame({col: 0 for col in self.power_columns}, index=[0]), self.power]).reset_index(drop=True)
        # Integrate the losses to obtain the cumulative power and the total power per cycle:
        self.cumulative_power, self.total_power_per_cycle = self.integrate_power() # Store cumulative power and total cumulative power per cycle
        # Store simulation time:
        try:
            with open(os.path.join(self.solution_folder, 'test_temporary', 'simulation_time.txt'), 'r') as f: 
                self.simulation_time = float(f.readline().strip())
        except:
            self.simulation_time = None # If the simulation time file does not exist, the simulation has not finished running.

        # Store the rest of the post-processing data:
        self.time = self.power['Time']
        self.instantaneous_temperature = self.load_standard_data(os.path.join(temp_file_path, 'temperature.txt'), 1, add_initial_zero=True)
        self.temperature = self.load_standard_data(os.path.join(temp_file_path, 'temperature.txt'), 2, add_initial_zero=True)
        self.I_transport = self.load_standard_data(os.path.join(temp_file_path, 'I_transport.txt'), 1)
        self.V_transport = self.load_standard_data(os.path.join(temp_file_path, 'V_transport.txt'), 1)
        self.hs_val = self.load_standard_data(os.path.join(temp_file_path, 'hs_val.txt'), 1)
        self.magn_fil = self.load_standard_data(os.path.join(temp_file_path, 'magn_fil.txt'), [1, 2, 3])
        self.magn_matrix = self.load_standard_data(os.path.join(temp_file_path, 'magn_matrix.txt'), [1, 2, 3])
        self.I = self.load_standard_data(os.path.join(temp_file_path, 'I.txt'), 1, len(self.time))
        self.V = self.load_standard_data(os.path.join(temp_file_path, 'V.txt'), 1, len(self.time))
        self.I_integral = self.load_standard_data(os.path.join(temp_file_path, 'I_integral.txt'), 1, len(self.time))
        self.I_abs_integral = self.load_standard_data(os.path.join(temp_file_path, 'I_abs_integral.txt'), 1, len(self.time))
        self.magnetic_energy_internal = self.load_standard_data(os.path.join(temp_file_path, 'magnetic_energy_internal.txt'), 1)
        self.Ip = self.load_standard_data(os.path.join(temp_file_path, 'Ip.txt'), 1, len(self.time))
        self.Vp = self.load_standard_data(os.path.join(temp_file_path, 'Vp.txt'), 1, len(self.time))


    def load_standard_data(self, file_path, columns, reshape = None, add_initial_zero = False):
        """
        There are many output .txt-files with similar format. This function loads the data from one of these files and returns it as a numpy array.
        If the file does not exist, None is returned without raising an error.
        """
        try:
            data = np.loadtxt(file_path, comments='#', usecols=columns)
            if reshape:
                data = data.reshape(-1, reshape).T
        except IOError:
            return None

        if add_initial_zero:
            if len(data.shape) == 1:
                data = np.insert(data, 0, 0)
            else:
                zeros = np.zeros((1, data.shape[1]))
                data = np.vstack((zeros, data))
        return data

    def retrieve_fiqusDataModel(self):
        """
        This function reads the YAML input-files for geometry, mesh and solve and stores them in three dictionaries which are returned.
        This function is to be called only once, when the object is created.
        """
        geometry_dataModel = load_yaml(os.path.join(self.geometry_folder, 'geometry.yaml'))
        mesh_dataModel = load_yaml(os.path.join(self.mesh_folder, 'mesh.yaml'))
        solution_dataModel = load_yaml(os.path.join(self.solution_folder, 'solve.yaml'))

        return geometry_dataModel, mesh_dataModel, solution_dataModel

    def integrate_power(self):
        """ 
        This function integrates the instantaneous power over time to obtain the cumulative power. 
        It also calculates the total cumulative power per cycle.
        The cumulative power is returned as a pandas dataframe and the total cumulative power per cycle is returned as a dictionary.
        """
        find_closest_idx = lambda arr, val: np.abs(arr - val).argmin() 

        t = np.array(self.power['Time'])
        t_final = t[-1]
        t_init = find_closest_idx(t, t_final/2)
        
        cumulative_power = pd.DataFrame(columns= self.power_columns)
        total_power_per_cycle = {}

        cumulative_power['Time'] = self.power["Time"]
        for column in self.power_columns[1:]:
            cumulative_power[column] = np.insert(integrate.cumulative_trapezoid(self.power[column], t), 0, 0)
            total_power_per_cycle[column] = 2 * (cumulative_power[column].iloc[-1]-cumulative_power[column].iloc[t_init]) # / (np.pi*matrix_radius**2 * loss_factor) # Why do we divide by pi*matrix_radius**2*loss_factor?

        return cumulative_power, total_power_per_cycle
    
    def plot_instantaneous_power(self, show:bool = True, title:str = "Power", save_plot:bool = False, save_folder_path:str = None, save_file_name:str = None, overwrite:bool = False):
        plt.figure()
        plt.plot(self.power['Time'], self.power[self.power_columns[1:]] , label = self.power_columns[1:])
        plt.xlabel('Time [s]')
        plt.ylabel('Power [W/m]')
        plt.legend()

        # Configure title:
        # Class attributes can be accessed by using '<< ... >>' in the title string.
        commands = re.findall('<<(.*?)>>', title)
        for c in commands:
            title = title.replace(f"<<{c}>>", str(eval('self.'+c)))
        plt.title(title)


        if save_plot: # Save the plot
            filePath = create_non_overwriting_filepath(save_folder_path, save_file_name, '.png', overwrite)
            plt.savefig(filePath)

        if show: 
            plt.show()
        else:
            plt.close()

class PlotPython:
    """
    This class loads and stores the data from the simulations specified in a csv file and can apply various postprocessing operations on the data. 
    The data from each simulation is saved as a SimulationData object which is subsequently stored in a list in this class.
    """
    def __init__(self, fdm, csv_filename = None, lossMap_gridData_folder = None, inputs_folder_path='', outputs_folder_path='') -> None:
        self.fdm = fdm
        self.inputs_folder_path = inputs_folder_path # This is the path to the folder where the input data is stored
        self.model_data_output_path = outputs_folder_path # This is the path to the folder where the model output data is stored (e.g. geometries)
        self.outputs_folder_path = os.path.join(outputs_folder_path, fdm.magnet.postproc.batch_postproc.output_folder) # This is the path to the folder where the postprocessed data is written
        

        if not os.path.exists(self.outputs_folder_path):
            os.makedirs(self.outputs_folder_path)

        if csv_filename is not None:
            try:
                self.input_csv = pd.read_csv(os.path.join(self.inputs_folder_path, f'{csv_filename}.csv')) # Read the csv file with the input data
            except:
                raise FileNotFoundError(f'No csv file with the name {fdm.magnet.postproc.batch_postproc.postProc_csv}.csv was found in the inputs folder.')
            
            self.simulation_collection = self.retrieve_simulation_data()

            self.avg_simulation_time = np.mean([sd.simulation_time for sd in self.simulation_collection])
            self.total_simulation_time = np.sum([sd.simulation_time for sd in self.simulation_collection])

            print('Number of simulations considered: ', len(self.simulation_collection)	)
            print('Average simulation time: ', self.avg_simulation_time, 's')
            print('Total simulation time: ', self.total_simulation_time, 's')



        elif lossMap_gridData_folder is not None:
            self.input_csv = None
            self.simulation_collection = None

            self.totalLoss_gridData = self.load_lossMap_gridData('TotalLoss', lossMap_gridData_folder)
            self.filamentLoss_gridData = self.load_lossMap_gridData('FilamentLoss', lossMap_gridData_folder)
            self.eddyLoss_gridData = self.load_lossMap_gridData('EddyLoss', lossMap_gridData_folder)
            self.couplingLoss_gridData = self.load_lossMap_gridData('CouplingLoss', lossMap_gridData_folder)
        else:
            raise ValueError('No input data specified. Either a csv file or a folder with loss map grid data must be provided.')
            
        
        

    def retrieve_simulation_data(self):
        """
        This function iterates over the input CSV-file (specifying which simulations to postprocess) and returns a list of SimulationData objects
        containing all the simulation data. If no CSV-file is specified, the data from the single simulation specified in the input YAML-file is returned.
        """
        if self.input_csv is not None:
            simulationCollection = []
            for index, row in self.input_csv.iterrows():
                if pd.isna(row['input.run.geometry']) and pd.isna(row['input.run.mesh']) and pd.isna(row['input.run.solution']):
                    continue
                geometry_name = 'Geometry_'+str(row['input.run.geometry'])
                mesh_name = 'Mesh_'+str(row['input.run.mesh'])

                if isinstance(row['input.run.solution'], float) and row['input.run.solution'].is_integer():
                    solution_name = 'Solution_'+str(int(row['input.run.solution']))
                else:
                    solution_name = 'Solution_'+str(row['input.run.solution'])

                # Check if the row refers to a valid simulation by checking if the solution folder exists:
                # solution_folder = os.path.join(os.getcwd(), 'tests', '_outputs', self.fdm.general.magnet_name, geometry_name, mesh_name, solution_name)
                solution_folder = os.path.join(self.model_data_output_path, geometry_name, mesh_name, solution_name)
                if os.path.exists(solution_folder): # If the solution folder exists, add the simulation to the simulationCollection
                    sd = SimulationData(self.model_data_output_path, geometry_name, mesh_name, solution_name)
                    if sd.simulation_time is not None: # Only add the simulation if it has finished running (and therefore has written the simulation time to a file)
                        simulationCollection.append(sd)
        else:
            simulationCollection = [SimulationData(self.model_data_output_path, 'Geometry_'+self.fdm.run.geometry, 'Mesh_'+self.fdm.run.mesh, 'Solution_'+self.fdm.run.solution)]
        
        return self.sort_simulationCollection(self.filter_simulationCollection(simulationCollection))

    def filter_simulationCollection(self, simulationCollection):
        """ 
        This function is used to filter the simulationCollection based on the filter criterion specified in the yaml input file.
        An example of a filter criterion is '<<solve.source_parameters.sine.frequency>> == 18', which will disregard all simulations with frequency != 18Hz.
        """
        if self.fdm.magnet.postproc.batch_postproc.filter.apply_filter:
            filter_criterion = self.fdm.magnet.postproc.batch_postproc.filter.filter_criterion
            class_params = re.findall('<<(.*?)>>', filter_criterion)
            for cp in class_params:
                filter_criterion = filter_criterion.replace(f"<<{cp}>>", 'sd.'+cp)
            filtering_function = eval(f'lambda sd: {filter_criterion}')
            return list(filter(filtering_function, simulationCollection))
        else:
            return simulationCollection
    
    def sort_simulationCollection(self, simulationCollection):
        """ 
        This function is used to sort the simulationCollection based on the sort key specified in the yaml input file.
        An example of a sort key is 'solve.source_parameters.sine.frequency', which will sort the simulations based on frequency.
        """
        if self.fdm.magnet.postproc.batch_postproc.sort.apply_sort:
            sorting_function = eval(f'lambda sd: sd.{self.fdm.magnet.postproc.batch_postproc.sort.sort_key}')
            return sorted(simulationCollection, key=sorting_function)
        else:
            return simulationCollection

    def lossMap_createGridData(self, lossType = 'TotalLoss', x_val_to_include = None, y_val_to_include = None):
        """
        This function creates the grid data needed for the loss map, based on the yaml input file.
        Given a collection of simulations it interpolates the loss data between the datapoints to a grid and returns the grid data.
        """
        lm = self.fdm.magnet.postproc.batch_postproc.loss_map

        # Extract data from simulation collection and normalize
        x_arr = np.array([eval('sd.'+lm.x_val)/lm.x_norm for sd in self.simulation_collection])
        y_arr = np.array([eval('sd.'+lm.y_val)/lm.y_norm for sd in self.simulation_collection])
        loss = np.array([sd.total_power_per_cycle[lossType]/lm.loss_norm for sd in self.simulation_collection])

        # Logarithmic scaling
        if lm.x_log: x_arr = np.log10(x_arr)
        if lm.y_log: y_arr = np.log10(y_arr) 
        if lm.loss_log: loss = np.log10(loss)

        x_arr_interpolated = np.linspace(min(x_arr), max(x_arr), lm.x_steps)
        y_arr_interpolated = np.linspace(min(y_arr), max(y_arr), lm.y_steps)
        # Insert specific values to the grid if they are not already included (useful for cross sections)
        if x_val_to_include is not None and x_val_to_include not in x_arr_interpolated: 
            x_arr_interpolated = np.insert(x_arr_interpolated, np.where(x_arr_interpolated > x_val_to_include)[0][0], x_val_to_include)
        if y_val_to_include is not None and y_val_to_include not in y_arr_interpolated:
            y_arr_interpolated = np.insert(y_arr_interpolated, np.where(y_arr_interpolated > y_val_to_include)[0][0], y_val_to_include)

        # Create grid
        X, Y = np.meshgrid(x_arr_interpolated, y_arr_interpolated, indexing='ij')
        gridPoints = np.c_[X.ravel(), Y.ravel()]
        dataPoints = np.c_[x_arr, y_arr]

        # Interpolate the simulation data onto the grid
        V = interpolate.griddata(
            dataPoints,
            loss,
            gridPoints,
            method='linear' # Cubic produces cleaner plots. Any incentive to go back to linear?
        ).reshape(X.shape)

        return X, Y, V, dataPoints
    
    def save_lossMap_gridData(self, save_folder_name = 'lossMap_gridData'):
        """
        This function calls the lossMap_createGridData function and saves the grid data.
        """
        lm = self.fdm.magnet.postproc.batch_postproc.loss_map

        lossTypes = ['TotalLoss', 'FilamentLoss', 'EddyLoss', 'CouplingLoss', 'CouplingLoss_dyn', 'TotalLoss_dyn'] # Only in the case dynamic correction is used, must be changed later
        # 1) Create a folder to store the output files
        gridData_folder_path = create_non_overwriting_filepath(self.outputs_folder_path, save_folder_name, '', self.fdm.run.overwrite)
        if not os.path.exists(gridData_folder_path): os.makedirs(gridData_folder_path)
        # 2) Create the grid data for each loss type and save it
        for lossType in lossTypes:
            X, Y, V, _ = self.lossMap_createGridData(lossType)
            if lm.x_log: X = np.power(10, X)
            if lm.y_log: Y = np.power(10, Y)
            if lm.loss_log: V = np.power(10, V)
            np.savetxt(os.path.join(gridData_folder_path, f'{lossType}.txt'), np.column_stack((X.ravel(), Y.ravel(), V.ravel())), delimiter=' ', header=f'{lm.x_val} {lm.y_val} {lossType}', comments='')

    def load_lossMap_gridData(self, lossType = 'TotalLoss', save_folder_name = 'lossMap_gridData'):
        """
        This function loads the grid data for a given loss type.
        """
        lm = self.fdm.magnet.postproc.batch_postproc.loss_map
        gridData_folder_path = os.path.join(self.inputs_folder_path, save_folder_name)
        
        if not os.path.exists(gridData_folder_path):
            raise FileNotFoundError(f'The folder {gridData_folder_path} does not exist.')

        X, Y, V = np.loadtxt(os.path.join(gridData_folder_path, f'{lossType}.txt'), unpack=True, skiprows=1)

        if lm.x_log: X = np.log10(X)
        if lm.y_log: Y = np.log10(Y)
        if lm.loss_log: V = np.log10(V)

        # Get the unique counts of X and Y
        unique_X = np.unique(X)
        unique_Y = np.unique(Y)

        # Reshape the data
        X = X.reshape((len(unique_X), len(unique_Y)))
        Y = Y.reshape((len(unique_X), len(unique_Y)))
        V = V.reshape((len(unique_X), len(unique_Y)))

        return X, Y, V
    
    # def add_value_to_gridData(self, X, Y, V, x_val = None, y_val = None):
    #     """
    #     This function adds a value to the grid data.
    #     Steps: 
    #         1) Revert the grid data to 1D arrays
    #         2) Add x or y value or both to the arrays
    #         3) Reshape the arrays back to grid data, interpolating the loss to the new grid points
    #     """
    #     gridPoints 
    
    def save_magnetization(self):
        """
        This function saves the magnetization data for all simulations in the simulation collection.
        """
        magnetization_folder_path = create_non_overwriting_filepath(self.outputs_folder_path, 'magnetization', '', self.fdm.run.overwrite)
        if not os.path.exists(magnetization_folder_path): os.makedirs(magnetization_folder_path)
        for sd in self.simulation_collection:
            magnetization = sd.magn_fil + sd.magn_matrix
            magnetization = np.c_[sd.time, magnetization]
            np.savetxt(os.path.join(magnetization_folder_path, f'magn_f{sd.solve.source_parameters.sine.frequency}_b{sd.solve.source_parameters.sine.field_amplitude}_I{sd.solve.source_parameters.sine.current_amplitude}.txt'), magnetization, delimiter=' ', header='t x y z', comments='')
        
            


        
    
    def lossMap_crossSection(self, slice_value, axis_to_cut = 'x'):
        """
        This function returns the data corresponding to a cross section of the loss map, for all loss types.
        Given an axis and a value, it sweeps the other axis for the closest value and returns the data.
        Example: Given slice value 0 and axis x, it returns the data for the cross section at x = 0.
        """

        lm = self.fdm.magnet.postproc.batch_postproc.loss_map
        if axis_to_cut == 'x':
            x_val_to_include = slice_value
            y_val_to_include = None
        elif axis_to_cut == 'y':
            x_val_to_include = None
            y_val_to_include = slice_value
        X, Y, V, dataPoints = self.lossMap_createGridData('TotalLoss', x_val_to_include, y_val_to_include)
        _,_,FilamentLoss, _ = self.lossMap_createGridData('FilamentLoss', x_val_to_include, y_val_to_include)
        _,_,EddyLoss, _ = self.lossMap_createGridData('EddyLoss', x_val_to_include, y_val_to_include)
        _,_,CouplingLoss, _ = self.lossMap_createGridData('CouplingLoss', x_val_to_include, y_val_to_include)


        if axis_to_cut == 'x':
            index = np.abs(X[:, 0] - slice_value).argmin()
            slice_vals = Y[index, :]

        elif axis_to_cut == 'y':
            index = np.abs(Y[0, :] - slice_value).argmin()
            slice_vals = X[:, index]

        # Extract the loss values for the constant frequency across all applied fields
        totalLoss = V[index, :] if axis_to_cut == 'x' else V[:, index]
        filamentLoss = FilamentLoss[index, :] if axis_to_cut == 'x' else FilamentLoss[:, index]
        eddyLoss = EddyLoss[index, :] if axis_to_cut == 'x' else EddyLoss[:, index]
        couplingLoss = CouplingLoss[index, :] if axis_to_cut == 'x' else CouplingLoss[:, index]

        return slice_vals, totalLoss, filamentLoss, eddyLoss, couplingLoss

    def plot_lossMap_crossSection(self):
        """
        This function calls the lossMap_crossSection function and plots the data it returns, which is the loss for all values of one axis, given a constant value of the other axis.
        """
        
        if is_latex_installed():
            plt.rcParams['text.usetex'] = True
            plt.rcParams['font.family'] = 'times'
            plt.rcParams['font.size'] = 20

        lm = self.fdm.magnet.postproc.batch_postproc.loss_map
        slice_value = lm.cross_section.cut_value
        axis_to_cut = lm.cross_section.axis_to_cut

        if (lm.x_log and axis_to_cut == 'x') or (lm.y_log and axis_to_cut == 'y'): 
            slice_value = np.log10(slice_value)

        slice_vals, totalLoss, filamentLoss, eddyLoss, couplingLoss = self.lossMap_crossSection(slice_value, axis_to_cut = axis_to_cut)

        def log_formatter(x, pos):
            """
                Format the tick labels on the plot.
            """
            return f"$10^{{{int(x)}}}$"

        # Plot the loss with respect to applied field for the constant frequency
        fig, ax = plt.subplots(figsize=(8, 6))
        ax.plot(slice_vals, totalLoss, label=f'Total Loss')
        ax.plot(slice_vals, filamentLoss, label=f'Filament Loss')
        ax.plot(slice_vals, eddyLoss, label=f'Eddy Loss')
        ax.plot(slice_vals, couplingLoss, label=f'Coupling Loss')

        tick_formatter = FuncFormatter(log_formatter)
        if lm.x_log and axis_to_cut == 'y' or lm.y_log and axis_to_cut == 'x': 
            ax.xaxis.set_major_formatter(tick_formatter)
        if lm.loss_log:
            ax.yaxis.set_major_formatter(tick_formatter)
        
        
        title = lm.cross_section.title.replace('<<cut_value>>', str(round(10**slice_value, 3)))
        ax.set_title(title)
        ax.set_xlabel(lm.ylabel if axis_to_cut == 'x' else lm.xlabel)
        ax.set_ylabel(lm.cross_section.ylabel)
        ax.legend()

        # np.savetxt(os.path.join(self.outputs_folder_path, 'lossMaps_cut_0p2T_0A.txt'), np.column_stack((10**slice_vals, 10**totalLoss, 10**eddyLoss, 10**couplingLoss, 10**filamentLoss)), delimiter=' ', header='f total eddy coupling filament', comments='')

        if lm.cross_section.save_plot:
            filePath = create_non_overwriting_filepath(self.outputs_folder_path, lm.cross_section.filename, '.png', self.fdm.run.overwrite)
            plt.savefig(filePath)

        if self.fdm.run.launch_gui: plt.show()

    def animate_lossMap_crossSection(self):
        """
        This function is similar to the plot_lossMap_crossSection function, but instead of plotting the loss for at a constant crossection, 
        it sweeps the crossection over a chosen axis and plots the loss for each crossection as an animation.
        """
        lm = self.fdm.magnet.postproc.batch_postproc.loss_map
        axis = lm.cross_section_sweep.axis_to_sweep

        X, Y, V, dataPoints = self.lossMap_createGridData('TotalLoss')
        x_vals = X[:, 0] # x-values from the loss map
        y_vals = Y[0, :] # y-values from the loss map

        if axis == 'x':
            A = np.zeros((lm.y_steps, 4, lm.x_steps))
            axis_to_sweep = x_vals
            constant_axis = y_vals
        elif axis == 'y':
            A = np.zeros((lm.x_steps, 4, lm.y_steps))
            axis_to_sweep = y_vals
            constant_axis = x_vals


        for i, val in enumerate(axis_to_sweep):
            _, totalLoss, filamentLoss, eddyLoss, couplingLoss = self.lossMap_crossSection(val, axis_to_cut = axis)
            A[:, 0, i] = totalLoss
            A[:, 1, i] = filamentLoss
            A[:, 2, i] = eddyLoss
            A[:, 3, i] = couplingLoss
        
        # Initialize the plot
        fig, ax = plt.subplots()
        lines = ax.plot(constant_axis, A[:, :, 0], lw=2, label = ['total Loss', 'filament Loss', 'eddy Loss', 'coupling Loss'])

        # Set plot limits and labels
        ax.set_xlim(constant_axis[0], constant_axis[-1])
        ax.set_ylim(np.min(A), np.max(A))
        ax.set_xlabel(lm.ylabel if axis == 'x' else lm.xlabel)
        ax.set_ylabel(lm.cross_section_sweep.ylabel)


        # Define the animation update function
        def update(frame):
            for i, line in enumerate(lines):
                line.set_ydata(A[:, i, frame])
            
            if axis == 'x':
                if lm.x_log: 
                    sweep_value = 10**x_vals[frame]
                else:
                    sweep_value = x_vals[frame]
            elif axis == 'y':
                if lm.y_log:
                    sweep_value = 10**y_vals[frame]
                else:
                    sweep_value = y_vals[frame]

            title = lm.cross_section_sweep.title.replace('<<sweep_value>>', str(round(sweep_value, 3)))
            ax.set_title(title)
            return lines,# line1, line2, line3, line4

        # Create the animation
        dt = 0.1
        ani = FuncAnimation(fig, update, frames=lm.x_steps if axis == 'x' else lm.y_steps, interval=dt*1000, blit=False)

        # Show the animation
        plt.legend()
        plt.grid()
        
        if lm.cross_section_sweep.save_plot:
            filepath = create_non_overwriting_filepath(folder_path=self.outputs_folder_path, base_name=lm.cross_section_sweep.filename, extension='.gif', overwrite=self.fdm.run.overwrite)
            ani.save(filepath, writer='imagemagick', fps=1/dt)

        if self.fdm.run.launch_gui: plt.show()

    def create_lossMap(self):
        """
        This function creates a loss map based on the inputs given in the loss_map section of the input file.
        The loss-map can be plotted and saved as a .png file.
        """
        lm = self.fdm.magnet.postproc.batch_postproc.loss_map

        if self.simulation_collection:
            X, Y, V, dataPoints = self.lossMap_createGridData(lm.loss_type)
        else:
            X, Y, V = self.totalLoss_gridData

        if is_latex_installed():
            plt.rcParams['text.usetex'] = True
            plt.rcParams['font.family'] = 'times'
            plt.rcParams['font.size'] = 20

        fig, ax = plt.subplots(figsize=(10,8))
    
        c = plt.pcolormesh(X, Y, V, shading='gouraud', cmap='plasma_r')
        c_min = min([np.ceil(np.min(V)) for V in [V]])
        c_max = max([np.floor(np.max(V)) for V in [V]])
        c_ticks = [int(val) for val in np.arange(c_min, c_max+1)]
        cont = plt.contour(X,Y,V, c_ticks, colors='k', linestyles='dashed')
        
        if lm.show_datapoints: 
            plt.scatter(dataPoints[:, 0], dataPoints[:, 1], s=50, edgecolors='k')

        
        if lm.show_loss_type_dominance_contour:
            sigmoid = lambda x: 1/(1+np.exp(-x))
            if self.simulation_collection:
                _, _, FilamentLoss, _ = self.lossMap_createGridData(lossType='FilamentLoss')
                _, _, CouplingLoss, _ = self.lossMap_createGridData(lossType='CouplingLoss')
                _, _, EddyLoss, _ = self.lossMap_createGridData(lossType='EddyLoss')
            # else:
            #     _, _, FilamentLoss = self.filamentLoss_gridData
            #     _, _, CouplingLoss = self.couplingLoss_gridData
            #     _, _, EddyLoss = self.eddyLoss_gridData
            fil_vs_coupling_loss = np.maximum(FilamentLoss, EddyLoss) - CouplingLoss
            fil_vs_eddy_loss = EddyLoss - np.maximum(FilamentLoss, CouplingLoss)
            plt.contour(X,Y,sigmoid(fil_vs_coupling_loss),[0.5], colors='k')
            plt.contour(X,Y,sigmoid(fil_vs_eddy_loss),[0.5], colors='k')
        
        cbar = fig.colorbar(c, ticks=c_ticks)#, labels=c_labels)
        # cbar.ax.set_xticks([-7, -6, -5, -4, -3, -2, -1, 0, 1])
        # cbar.ax.set_yticklabels([r'$10^{-7}$', r'$10^{-6}$', r'$10^{-5}$', r'$10^{-4}$', r'$10^{-3}$', r'$10^{-2}$', r'$10^{-1}$', r'$10^0$', r'$10^1$'])
        cbar.ax.set_yticklabels([f"$10^{{{val}}}$" for val in c_ticks])
        # plt.grid(alpha=0.5)
        # plt.title(lm.title)
        # plt.xlabel(lm.xlabel)
        # plt.ylabel(lm.ylabel)
        plt.title(r'Loss per cycle (J/m)')
        plt.xlabel(r'Frequency $f$ (Hz)')
        plt.ylabel(r'Field amplitude $b$ (T)')


        # plt.annotate(r'Coupling', (np.log10(1.0), np.log10(0.007)), color='white')
        # plt.annotate(r'Filament', (np.log10(0.012), np.log10(0.74)), color='white')
        # plt.annotate(r'(uncoupled)', (np.log10(0.012), np.log10(0.55)), color='white')
        # plt.annotate(r'Filament', (np.log10(45), np.log10(0.38)), color='white')
        # plt.annotate(r'(coupled)', (np.log10(45), np.log10(0.28)), color='white')
        # plt.annotate(r'Eddy', (np.log10(2000), np.log10(0.03)), color='white')

        # ax.plot(np.log10(0.03), np.log10(0.2),  'o', color='white')#, xytext=(np.log10(0.03), np.log10(0.12)), arrowprops=dict(facecolor='black', shrink=0.02))
        # ax.plot(np.log10(30),   np.log10(1),    'o', color='white')#, xytext=(np.log10(40), np.log10(0.8)), arrowprops=dict(facecolor='black', shrink=0.02))
        # ax.plot(np.log10(3),    np.log10(0.2),  'o', color='white')#, xytext=(np.log10(2), np.log10(0.2)), arrowprops=dict(facecolor='black', shrink=0.02))
        # ax.plot(np.log10(5000), np.log10(0.2),  'o', color='white')#, xytext=(np.log10(5000), np.log10(0.1)), arrowprops=dict(facecolor='black', shrink=0.02))

        # ax.annotate('(a)', xy=(np.log10(0.03), np.log10(0.2)), xycoords='data', ha='right', va='bottom', fontsize=20, color='white')
        # ax.annotate('(b)', xy=(np.log10(3), np.log10(0.2)), xycoords='data', ha='right', va='bottom', fontsize=20, color='white')
        # ax.annotate('(c)', xy=(np.log10(30), np.log10(1)), xycoords='data', ha='right', va='bottom', fontsize=20, color='white')
        # ax.annotate('(d)', xy=(np.log10(5000), np.log10(0.2)), xycoords='data', ha='right', va='bottom', fontsize=20, color='white')

        # Define custom tick labels for x-axis
        x_min_log = int(np.log10(min([eval('sd.'+lm.x_val) for sd in self.simulation_collection])))
        x_max_log = int(np.log10(max([eval('sd.'+lm.x_val) for sd in self.simulation_collection])))
        x = np.arange(x_min_log, x_max_log+1)
        # Create a list of minor ticks
        minor_x_labels = []
        # 1) Add the ticks from x_min_log to ceil(x_min_log) to the minor_x_test list
        new_ticks = np.linspace(10.0**np.floor(x_min_log), 10.0**np.ceil(x_min_log), 10)[:-1]
        new_ticks = np.unique(new_ticks[new_ticks >= 10.0**x_min_log])
        minor_x_labels.extend(new_ticks)
        # 2) Add the ticks from ceil(x_min_log) to floor(x_max_log) to the minor_x_test list
        for x_val in x:
            new_ticks = np.linspace(10.0**x_val, 10.0**(x_val+1), 10)[1:-1]
            if x_val == x[-1]:
                new_ticks = new_ticks[new_ticks <= 10.0**x_max_log]
            minor_x_labels.extend(new_ticks)
        minor_x = [np.log10(val) for val in minor_x_labels]
        
        new_x_labels = [f"$10^{{{val}}}$" for val in x]
        plt.xticks(x, new_x_labels)
        plt.xticks(minor_x, minor=True)

        # Define custom tick labels for y-axis
        y_min_log = np.log10(min([eval('sd.'+lm.y_val) for sd in self.simulation_collection]))
        y_max_log = np.log10(max([eval('sd.'+lm.y_val) for sd in self.simulation_collection]))
        y = np.arange(np.ceil(y_min_log), np.floor(y_max_log)+1)
        # Create a list of minor ticks
        minor_y_labels = []
        # 1) Add the ticks from y_min_log to ceil(y_min_log) to the minor_y_test list
        new_ticks = np.linspace(10.0**np.floor(y_min_log), 10.0**np.ceil(y_min_log), 10)[:-1]
        new_ticks = np.unique(new_ticks[new_ticks >= 10.0**y_min_log])
        minor_y_labels.extend(new_ticks)
        # 2) Add the ticks from ceil(y_min_log) to floor(y_max_log) to the minor_y_test list
        for y_val in y:
            new_ticks = np.linspace(10.0**y_val, 10.0**(y_val+1), 10)[1:-1]
            if y_val == y[-1]:
                new_ticks = new_ticks[new_ticks <= 10.0**y_max_log]
            minor_y_labels.extend(new_ticks)

        new_y_labels = [f"$10^{{{int(val)}}}$" for val in y]
        minor_y = [np.log10(val) for val in minor_y_labels]
        plt.yticks(y, new_y_labels)
        plt.yticks(minor_y, minor=True)

        # plt.savefig('C:/Users/jdular/cernbox/Documents/Reports/CERN_Reports/linkedFluxPaper/fig/loss_map_54fil_noI.pdf', bbox_inches='tight')


        if lm.save_plot:
            filePath = create_non_overwriting_filepath(self.outputs_folder_path, lm.filename, '.pdf', self.fdm.run.overwrite)
            plt.savefig(filePath, bbox_inches='tight')

        if self.fdm.run.launch_gui: plt.show()


    def plot2d(self):  
        """
        This function is used to create a 2d plot. It is supposed to be flexible and work for various kinds of plots one may want to create.
        """
        if is_latex_installed():
            plt.rcParams['text.usetex'] = True
            plt.rcParams['font.family'] = 'times'
            # plt.rcParams['font.size'] = 20

        # Create the title (or titles if combined_plot is False)
        title = self.fdm.magnet.postproc.batch_postproc.plot2d.title
        if self.fdm.magnet.postproc.batch_postproc.plot2d.combined_plot:
            sd = self.simulation_collection[0]
            commands = re.findall('<<(.*?)>>', title)
            for c in commands:
                title = title.replace(f"<<{c}>>", str(eval('sd.'+c)))
        else:
            titles = []
            for sd in self.simulation_collection:
                commands = re.findall('<<(.*?)>>', title)
                title_i = title
                for c in commands:
                    title_i = title_i.replace(f"<<{c}>>", str(eval('sd.'+c)))
                titles.append(title_i)

        # Create the labels
        label_list = self.fdm.magnet.postproc.batch_postproc.plot2d.labels
        labels = np.zeros((len(self.simulation_collection), len(label_list)), dtype=object)
        for i, sd in enumerate(self.simulation_collection):
            simulation_labels = []
            for l in label_list:
                commands = re.findall('<<(.*?)>>', l)
                for c in commands:
                    l = l.replace(f"<<{c}>>", str(eval('sd.'+c)))

                simulation_labels.append(l)
            labels[i, :] = simulation_labels

        # colors = plt.cm.get_cmap('magma').resampled(len(self.simulation_collection)).colors
        colors = plt.cm.get_cmap('viridis').resampled(len(self.simulation_collection)).colors

        # Load the x-values:
        x_val = self.fdm.magnet.postproc.batch_postproc.plot2d.x_val
        commands = re.findall('<<(.*?)>>', x_val)
        for c in commands:
            x_val = x_val.replace(f"<<{c}>>", str('sd.'+c))
        x_arr = np.array([eval(x_val) for sd in self.simulation_collection], dtype=object)

        # Load the y-values:
        y_vals = self.fdm.magnet.postproc.batch_postproc.plot2d.y_vals
        y_arr = np.zeros((len(self.simulation_collection), len(y_vals)), dtype=object)
        for i, sd in enumerate(self.simulation_collection):
            for j, y_val in enumerate(y_vals):
                commands = re.findall('<<(.*?)>>', y_val)
                for c in commands:
                    y_val = y_val.replace(f"<<{c}>>", str('sd.'+c))
                y_arr[i, j] = eval(y_val)

        # data = np.column_stack((x_arr, y_arr))
        # np.savetxt(os.path.join(self.outputs_folder_path, self.fdm.magnet.postproc.batch_postproc.plot2d.filename+'.txt'), data, delimiter=' ', header='f total eddy coupling filament', comments='')

        
        

        # Plot and save the data:
        if not self.fdm.magnet.postproc.batch_postproc.plot2d.combined_plot and self.fdm.magnet.postproc.batch_postproc.plot2d.save_plot:  
            # Create a folder to save the plots if combined_plot is False and save_plot is True:  
            filename = self.fdm.magnet.postproc.batch_postproc.plot2d.filename
            folder_path = create_non_overwriting_filepath(self.outputs_folder_path, filename, '', self.fdm.run.overwrite)
            if not os.path.exists(folder_path): os.makedirs(folder_path)

        # Check if the y-values are all floats:
        y_is_float = np.all(np.apply_along_axis(lambda arr: np.all(np.vectorize(isinstance)(arr, float)), axis=1, arr=y_arr))
        # If they are all floats (not arrays), we can make a single plot, spanning all simulations, instead of one plot per simulation.
        if y_is_float:
            for column in range(y_arr.shape[1]):
                plt.plot(x_arr, y_arr[:, column], label=label_list[column])
            if self.fdm.magnet.postproc.batch_postproc.plot2d.legend:
                plt.legend()
            plt.grid()
            plt.xlabel(self.fdm.magnet.postproc.batch_postproc.plot2d.xlabel)
            plt.ylabel(self.fdm.magnet.postproc.batch_postproc.plot2d.ylabel)
            plt.title(title)
            if self.fdm.magnet.postproc.batch_postproc.plot2d.x_log:
                plt.xscale('log')
            if self.fdm.magnet.postproc.batch_postproc.plot2d.y_log:
                plt.yscale('log')

            if self.fdm.magnet.postproc.batch_postproc.plot2d.save_plot:
                filename = self.fdm.magnet.postproc.batch_postproc.plot2d.filename
                filePath = create_non_overwriting_filepath(self.outputs_folder_path, filename, '.png', self.fdm.run.overwrite)
                plt.savefig(filePath)

            
        else:
            for i, (x, y_vals_per_sim, labels_per_sim, color) in enumerate(zip(x_arr, y_arr, labels, colors)):
                if self.fdm.magnet.postproc.batch_postproc.plot2d.combined_plot:
                    # If combined_plot is true, plot all the data in the same figure:
                    # x_sin = 2*np.sin(2*np.pi*x/x.iloc[-1]) #temporary for missing hs_val
                    for y, label in zip(y_vals_per_sim, labels_per_sim):
                        # plt.plot(x_sin, x_sin+1.6*y, self.fdm.magnet.postproc.batch_postproc.plot2d.linestyle, label=label, color = color) #temporary for missing hs_val
                        plt.plot(x, y, self.fdm.magnet.postproc.batch_postproc.plot2d.linestyle, label=label, color = color)
                else:
                    # If combined_plot is false, plot data from each simulation in a separate figure:
                    plt.figure()
                    for y, label in zip(y_vals_per_sim, labels_per_sim):
                        plt.plot(x, y, self.fdm.magnet.postproc.batch_postproc.plot2d.linestyle, label=label)

                # Set the plot options
                if self.fdm.magnet.postproc.batch_postproc.plot2d.legend:
                    plt.legend()
                plt.grid()
                plt.xlabel(self.fdm.magnet.postproc.batch_postproc.plot2d.xlabel)
                plt.ylabel(self.fdm.magnet.postproc.batch_postproc.plot2d.ylabel)
                plt.title(title if self.fdm.magnet.postproc.batch_postproc.plot2d.combined_plot else titles[i])


                if not self.fdm.magnet.postproc.batch_postproc.plot2d.combined_plot and self.fdm.magnet.postproc.batch_postproc.plot2d.save_plot:
                    # If combined_plot is false we expect a lot of plots and save them all into a folder:
                    filename = self.fdm.magnet.postproc.batch_postproc.plot2d.filename
                    plt.savefig(os.path.join(folder_path, filename+f"_{i}.png"))

            if self.fdm.magnet.postproc.batch_postproc.plot2d.combined_plot and self.fdm.magnet.postproc.batch_postproc.plot2d.save_plot:
                # If combined_plot is true we expect only one plot and save it in the main folder:
                filename = self.fdm.magnet.postproc.batch_postproc.plot2d.filename
                filePath = create_non_overwriting_filepath(self.outputs_folder_path, filename, '.png', self.fdm.run.overwrite)
                plt.savefig(filePath, dpi=300)

        if self.fdm.run.launch_gui: plt.show()

    



from pydantic import BaseModel, Field
from typing import Union, Literal, Optional

# ------------------- Jc fits ---------------------------#
class ConstantJc(BaseModel):
    """
    Level 3: Class for setting constant Jc
    """

    type: Literal["Constant Jc"]
    Jc_constant: Optional[float] = None  # [A/m^2]


class Ic_A_NbTi(BaseModel):
    """
    Level 3: Class for setting IcNbTi fit
    """

    type: Literal["Ic_A_NbTi"]
    Jc_5T_4_2K: Optional[float] = None  # [A/m^2]


class Bottura(BaseModel):
    """
    Level 3: Class for setting Bottura fit
    """

    type: Literal["Bottura"]
    Tc0_Bottura: Optional[float] = None  # [K]
    Bc20_Bottura: Optional[float] = None  # [T]
    Jc_ref_Bottura: Optional[float] = None  # [A/m^2]
    C0_Bottura: Optional[float] = None  # [-]
    alpha_Bottura: Optional[float] = None  # [-]
    beta_Bottura: Optional[float] = None  # [-]
    gamma_Bottura: Optional[float] = None  # [-]


class CUDI1(BaseModel):
    """
    Level 3: Class for Nb-Ti fit based on "Fit 1" in CUDI manual
    """

    type: Literal["CUDI1"]
    Tc0_CUDI1: Optional[float] = None  # [K]
    Bc20_CUDI1: Optional[float] = None  # [T]
    C1_CUDI1: Optional[float] = None  # [A]
    C2_CUDI1: Optional[float] = None  # [A/T]


class CUDI3(BaseModel):
    """
    Level 3: Class for Nb-Ti fit based on "Fit 3" in CUDI manual
    """

    type: Literal["CUDI3"]
    Tc0_CUDI3: Optional[float] = None  # [K]
    Bc20_CUDI3: Optional[float] = None  # [T]
    c1_CUDI3: Optional[float] = None  # [-]
    c2_CUDI3: Optional[float] = None  # [-]
    c3_CUDI3: Optional[float] = None  # [-]
    c4_CUDI3: Optional[float] = None  # [-]
    c5_CUDI3: Optional[float] = None  # [-]
    c6_CUDI3: Optional[float] = None  # [-]


class Summers(BaseModel):
    """
    Level 3: Class for cable Summer's Nb3Sn fit
    """

    type: Literal["Summers"]
    Tc0_Summers: Optional[float] = None  # [K]
    Bc20_Summers: Optional[float] = None  # [T]
    Jc0_Summers: Optional[float] = None  # [A*T^0.5/m^2]


class Bordini(BaseModel):
    """
    Level 3: Class for cable Bordini's Nb3Sn fit
    """

    type: Literal["Bordini"]
    Tc0_Bordini: Optional[float] = None  # [K]
    Bc20_Bordini: Optional[float] = None  # [T]
    C0_Bordini: Optional[float] = None  # [A*T/m^2]
    alpha_Bordini: Optional[float] = None  # [-]


class BSCCO_2212_LBNL(BaseModel):
    """
    Level 3: Class for cable Bi-2212 fit developed in LBNL
    """

    # only ad-hoc fit [T. Shen, D. Davis, E. Ravaioli with LBNL, Berkeley, CA]
    type: Literal["BSCCO_2212_LBNL"]
    f_scaling_Jc_BSCCO2212: Optional[float] = None  # [-] used for the ad-hoc fit


# ------------------- Cable types ---------------------------#
class Mono(BaseModel):
    """
    Mono cable type: This is basically type of cable consisting of one strand - not really a cable
    """

    type: Literal["Mono"]
    bare_cable_width: Optional[float] = None
    bare_cable_height_low: Optional[float] = None
    bare_cable_height_high: Optional[float] = None
    bare_cable_height_mean: Optional[float] = None
    th_insulation_along_width: Optional[float] = None
    th_insulation_along_height: Optional[float] = None
    # Fractions given with respect to the insulated conductor
    f_superconductor: Optional[float] = None
    f_stabilizer: Optional[float] = None  # (related to CuFraction in ProteCCT)
    f_insulation: Optional[float] = None
    f_inner_voids: Optional[float] = None
    f_outer_voids: Optional[float] = None
    # Available materials depend on the component and on the selected program
    material_insulation: Optional[str] = None
    material_inner_voids: Optional[str] = None
    material_outer_voids: Optional[str] = None


class Rutherford(BaseModel):
    """
    Rutherford cable type: for example LHC MB magnet cable
    """

    type: Literal["Rutherford"]
    n_strands: Optional[int] = None
    n_strand_layers: Optional[int] = None
    n_strands_per_layers: Optional[int] = None
    bare_cable_width: Optional[float] = None
    bare_cable_height_low: Optional[float] = None
    bare_cable_height_high: Optional[float] = None
    bare_cable_height_mean: Optional[float] = None
    th_insulation_along_width: Optional[float] = None
    th_insulation_along_height: Optional[float] = None
    width_core: Optional[float] = None
    height_core: Optional[float] = None
    strand_twist_pitch: Optional[float] = None
    strand_twist_pitch_angle: Optional[float] = None
    Rc: Optional[float] = None
    Ra: Optional[float] = None
    # Fractions given with respect to the insulated conductor
    f_superconductor: Optional[float] = None
    f_stabilizer: Optional[float] = None  # (related to CuFraction in ProteCCT)
    f_insulation: Optional[float] = None
    f_inner_voids: Optional[float] = None
    f_outer_voids: Optional[float] = None
    f_core: Optional[float] = None
    # Available materials depend on the component and on the selected program
    material_insulation: Optional[str] = None
    material_inner_voids: Optional[str] = None
    material_outer_voids: Optional[str] = None
    material_core: Optional[str] = None


class Ribbon(BaseModel):
    """
    Mono cable type: This is basically type of cable consisting of one strand - not really a cable
    """

    type: Literal["Ribbon"]
    n_strands: Optional[int] = (
        None  # This defines the number of "strands" in the ribbon cable, which are physically glued but electrically in series
    )
    bare_cable_width: Optional[float] = None
    bare_cable_height_low: Optional[float] = None
    bare_cable_height_high: Optional[float] = None
    bare_cable_height_mean: Optional[float] = None
    th_insulation_along_width: Optional[float] = (
        None  # This defines the thickness of the insulation around each strand (DIFFERENT FROM ROXIE CADATA FILE)
    )
    th_insulation_along_height: Optional[float] = (
        None  # This defines the thickness of the insulation around each strand (DIFFERENT FROM ROXIE CADATA FILE)
    )
    # Fractions given with respect to the insulated conductor
    f_superconductor: Optional[float] = None
    f_stabilizer: Optional[float] = None  # (related to CuFraction in ProteCCT)
    f_insulation: Optional[float] = None
    f_inner_voids: Optional[float] = None
    f_outer_voids: Optional[float] = None
    f_core: Optional[float] = None
    # Available materials depend on the component and on the selected program
    material_insulation: Optional[str] = None
    material_inner_voids: Optional[str] = None
    material_outer_voids: Optional[str] = None
    material_core: Optional[str] = None


# ------------------- Conductors ---------------------------#

# class MaterialSuperconductor(BaseModel):
#     """
#     Level 3: Class for strand superconductor material parameters
#     """
#     material: Optional[str] = Field(default=None, description="Material of the superconductor. E.g. NbTi, Nb3Sn, etc.")
#     n_value: Optional[float] = Field(default=None, description="n value of the superconductor (for power law fit).")
#     ec: Optional[float] = Field(default=None, description="Critical electric field of the superconductor.")
#     Cv_material: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the superconductor.")

# class MaterialStabilizer(BaseModel):
#     """
#     Level 3: Class for strand stabilizer material parameters
#     """

#     rho_material: Optional[Union[str, float]] = Field(default=None, description="Material function for resistivity of the stabilizer. Constant resistivity can be given as float.")
#     RRR: Optional[float] = Field(default=None, description="Residual resistivity ratio of the stabilizer.")
#     T_ref_RRR_high: Optional[float] = Field(default=None, description="Upper reference temperature for RRR measurements.")
#     T_ref_RRR_low: Optional[float] = Field(default=None, description="Lower reference temperature for RRR measurements.")
#     k_material: Optional[Union[str, float]] = Field(default=None, description="Thermal conductivity of the stabilizer.")
#     Cv_material: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the stabilizer.")


class Round(BaseModel):
    """
    Level 2: Class for strand parameters
    """

    type: Literal["Round"]
    fil_twist_pitch: Optional[float] = None # Strand twist pitch
    diameter: Optional[float] = None  # ds_inGroup (LEDET), DConductor (BBQ), DStrand (ProteCCT)
    diameter_core:  Optional[float] = None  # dcore_inGroup (LEDET)
    diameter_filamentary: Optional[float] = None  # dfilamentary_inGroup (LEDET)
    filament_diameter: Optional[float] = None  # df_inGroup (LEDET)
    number_of_filaments: Optional[int] = None  # nf_inGroup (LEDET)
    f_Rho_effective: Optional[float] = None
    Cu_noCu_in_strand: Optional[float] = None

    # -- Superconductor parameters -- #
    material_superconductor: Optional[str] = Field(default=None, description="Material of the superconductor. E.g. Nb-Ti, Nb3Sn, etc.")
    n_value_superconductor: Optional[float] = Field(default=None, description="n value of the superconductor (for power law fit).")
    ec_superconductor: Optional[float] = Field(default=None, description="Critical electric field of the superconductor.")
    k_material_superconductor: Optional[Union[str, float]] = Field(default=None, description="Thermal conductivity of the superconductor.")
    Cv_material_superconductor: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the superconductor.")
    # -- Stabilizer parameters -- #
    material_stabilizer: Optional[str] = Field(default=None, description="Material of the stabilizer.")
    rho_material_stabilizer: Optional[Union[str, float]] = Field(default=None, description="Material function for resistivity of the stabilizer. Constant resistivity can be given as float.")
    RRR: Optional[float] = Field(default=None, description="Residual resistivity ratio of the stabilizer.")
    T_ref_RRR_high: Optional[float] = Field(default=None, description="Upper reference temperature for RRR measurements.")
    T_ref_RRR_low: Optional[float] = Field(default=None, description="Lower reference temperature for RRR measurements.")
    k_material_stabilizer: Optional[Union[str, float]] = Field(default=None, description="Thermal conductivity of the stabilizer.")
    Cv_material_stabilizer: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the stabilizer.")

    # superconductor: MaterialSuperconductor = MaterialSuperconductor()
    # stabilizer: MaterialStabilizer = MaterialStabilizer()


class Rectangular(BaseModel):
    """
    Level 2: Class for strand parameters
    """

    type: Literal["Rectangular"]
    bare_width: Optional[float] = None
    bare_height: Optional[float] = None
    Cu_noCu_in_strand: Optional[float] = None
    filament_diameter: Optional[float] = None  # df_inGroup (LEDET)
    f_Rho_effective: Optional[float] = None
    fil_twist_pitch: Optional[float] = None
    bare_corner_radius: Optional[float] = None

    # -- Superconductor parameters -- #
    material_superconductor: Optional[str] = Field(default=None, description="Material of the superconductor. E.g. NbTi, Nb3Sn, etc.")
    n_value_superconductor: Optional[float] = Field(default=None, description="n value of the superconductor (for power law fit).")
    ec_superconductor: Optional[float] = Field(default=None, description="Critical electric field of the superconductor.")
    k_material_superconductor: Optional[Union[str, float]] = Field(default=None, description="Thermal conductivity of the superconductor.")
    Cv_material_superconductor: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the superconductor.")
    # -- Stabilizer parameters -- #
    material_stabilizer: Optional[str] = Field(default=None, description="Material of the stabilizer.")
    rho_material_stabilizer: Optional[Union[str, float]] = Field(default=None, description="Material function for resistivity of the stabilizer. Constant resistivity can be given as float.")
    RRR: Optional[float] = Field(default=None, description="Residual resistivity ratio of the stabilizer.")
    T_ref_RRR_high: Optional[float] = Field(default=None, description="Upper reference temperature for RRR measurements.")
    T_ref_RRR_low: Optional[float] = Field(default=None, description="Lower reference temperature for RRR measurements.")
    k_material_stabilizer: Optional[Union[str, float]] = Field(default=None, description="Thermal conductivity of the stabilizer.")
    Cv_material_stabilizer: Optional[Union[str, float]] = Field(default=None, description="Material function for specific heat of the stabilizer.")

    # superconductor: MaterialSuperconductor = MaterialSuperconductor()
    # stabilizer: MaterialStabilizer = MaterialStabilizer()
    


# ------------------- Conductors ---------------------------#


class Conductor(BaseModel):
    """
    Level 1: Class for conductor parameters
    """

    version: Optional[str] = None
    case: Optional[str] = None
    state: Optional[str] = None
    cable: Union[Rutherford, Mono, Ribbon] = {
        "type": "Rutherford"
    }  # TODO: Busbar, Rope, Roebel, CORC, TSTC, CICC
    strand: Union[Round, Rectangular] = {"type": "Round"}  # TODO: Tape, WIC
    Jc_fit: Union[ConstantJc, Bottura, CUDI1, CUDI3, Summers, Bordini, BSCCO_2212_LBNL, Ic_A_NbTi] = {
        "type": "CUDI1"
    }  # TODO: CUDI other numbers? , Roxie?



from pydantic import BaseModel
from typing import List, Optional


class Terminal(BaseModel):
    vol_st: Optional[List[int]] = None  # volume number for terminal in for straightening
    surf_st: Optional[List[int]] = None  # surface number for terminal in for straightening
    vol_et: Optional[List[int]] = None  # volume number for terminal in for extending
    surf_et: Optional[List[int]] = None  # surface number for terminal in for extending
    lc_st: Optional[List[List[List[int]]]] = None  # line connections for straightening terminals
    lc_et: Optional[List[List[List[int]]]] = None  # line connections for extending terminals
    z_air: Optional[float] = None
    z_add: Optional[float] = None
    ndpterms: Optional[List[int]] = None  # number of divisions per terminal


class Winding(BaseModel):
    names: Optional[List[str]] = None  # name to use in gmsh and getdp
    t_in: Terminal = Terminal()     # Terminal in
    t_out: Terminal = Terminal()    # Terminal in


class WindingsInformation(BaseModel):
    magnet_name: Optional[str] = None
    windings_avg_length: Optional[float] = None
    windings: Winding = Winding()
    w_names: Optional[List[str]] = None
    f_names: Optional[List[str]] = None
    formers: Optional[List[str]] = None
    air: Optional[str] = None


class SpliterBrep(BaseModel):  # Brep file model splitter data
    magnet_name: Optional[str] = None
    file_name: Optional[str] = None           # full file name for the brep file
    vol_firsts: Optional[List[int]] = None    # list of first volumes for the partitioned model
    vol_lasts: Optional[List[int]] = None      # list of last volumes for the partitioned model



from pydantic import BaseModel
from typing import Dict, List, Optional


class Coord(BaseModel):
    """
    Class for coordinates
    """

    x: Optional[float] = None
    y: Optional[float] = None
    z: Optional[float] = None


class Roll(BaseModel):
    """
    Class for roll2 transformation
    """

    coor: Coord = Coord()
    alph: Optional[float] = None


class HyperHole(BaseModel):
    """
    Class for hyper holes
    """

    areas: List[str] = []


class HyperArea(BaseModel):
    """
    Class for hyper areas
    """

    material: Optional[str] = None
    lines: List[str] = []


class HyperLine(BaseModel):
    """
    Class for hyper lines: lines, arcs, elliptic arcs, circles
    """

    type: Optional[str] = None
    kp1: Optional[str] = None
    kp2: Optional[str] = None
    kp3: Optional[str] = None
    arg1: Optional[float] = None
    arg2: Optional[float] = None
    elements: Optional[int] = None


class CondPar(BaseModel):
    """
    Class for conductor parameters
    """

    wInsulNarrow: Optional[float] = None
    wInsulWide: Optional[float] = None
    dFilament: Optional[float] = None
    dstrand: Optional[float] = None
    fracCu: Optional[float] = None
    fracSc: Optional[float] = None
    RRR: Optional[float] = None
    TupRRR: Optional[float] = None
    Top: Optional[float] = None
    Rc: Optional[float] = None
    Ra: Optional[float] = None
    fRhoEff: Optional[float] = None
    lTp: Optional[float] = None
    wBare: Optional[float] = None
    hInBare: Optional[float] = None
    hOutBare: Optional[float] = None
    noOfStrands: Optional[int] = None
    noOfStrandsPerLayer: Optional[int] = None
    noOfLayers: Optional[int] = None
    lTpStrand: Optional[float] = None
    wCore: Optional[float] = None
    hCore: Optional[float] = None
    thetaTpStrand: Optional[float] = None
    degradation: Optional[float] = None
    C1: Optional[float] = None
    C2: Optional[float] = None
    fracHe: Optional[float] = None
    fracFillInnerVoids: Optional[float] = None
    fracFillOuterVoids: Optional[float] = None


class Conductor(BaseModel):
    """
    Class for conductor type
    """

    conductorType: Optional[int] = None
    cableGeom: Optional[str] = None
    strand: Optional[str] = None
    filament: Optional[str] = None
    insul: Optional[str] = None
    trans: Optional[str] = None
    quenchMat: Optional[str] = None
    T_0: Optional[float] = None
    comment: Optional[str] = None
    parameters: CondPar = CondPar()


class Cable(BaseModel):
    """
    Class for cable parameters
    """

    height: Optional[float] = None
    width_i: Optional[float] = None
    width_o: Optional[float] = None
    ns: Optional[float] = None
    transp: Optional[float] = None
    degrd: Optional[float] = None
    comment: Optional[str] = None


class Quench(BaseModel):
    """
    Class for quench parameters
    """

    SCHeatCapa: Optional[float] = None
    CuHeatCapa: Optional[float] = None
    CuThermCond: Optional[float] = None
    CuElecRes: Optional[float] = None
    InsHeatCapa: Optional[float] = None
    InsThermCond: Optional[float] = None
    FillHeatCapa: Optional[float] = None
    He: Optional[float] = None
    comment: Optional[str] = None


class Transient(BaseModel):
    """
    Class for transient parameters
    """

    Rc: Optional[float] = None
    Ra: Optional[float] = None
    filTwistp: Optional[float] = None
    filR0: Optional[float] = None
    fil_dRdB: Optional[float] = None
    strandfillFac: Optional[float] = None
    comment: Optional[str] = None


class Strand(BaseModel):
    """
    Class for strand parameters
    """

    diam: Optional[float] = None
    cu_sc: Optional[float] = None
    RRR: Optional[float] = None
    Tref: Optional[float] = None
    Bref: Optional[float] = None
    Jc_BrTr: Optional[float] = None
    dJc_dB: Optional[float] = None
    comment: Optional[str] = None


class Filament(BaseModel):
    """
    Class for filament parameters
    """

    fildiao: Optional[float] = None
    fildiai: Optional[float] = None
    Jc_fit: Optional[str] = None
    fit: Optional[str] = None
    comment: Optional[str] = None


class Insulation(BaseModel):
    """
    Class for insulation parameters
    """

    radial: Optional[float] = None
    azimut: Optional[float] = None
    comment: Optional[str] = None


class Block(BaseModel):
    """
    Class for block list
    """

    type: Optional[int] = None
    nco: Optional[int] = None
    radius: Optional[float] = None
    phi: Optional[float] = None
    alpha: Optional[float] = None
    current: Optional[float] = None
    condname: Optional[str] = None
    n1: Optional[int] = None
    n2: Optional[int] = None
    imag: Optional[int] = None
    turn: Optional[float] = None
    coil: Optional[int] = None
    pole: Optional[int] = None
    layer: Optional[int] = None
    winding: Optional[int] = None
    shift2: Coord = Coord()
    roll2: Roll = Roll()


class Group(BaseModel):
    """
    Class for group list
    """

    symm: Optional[int] = None
    typexy: Optional[int] = None
    blocks: List[int] = []  # map


class Trans(BaseModel):
    """
    Class for transformation list
    """

    x: Optional[float] = None
    y: Optional[float] = None
    alph: Optional[float] = None
    bet: Optional[float] = None
    string: Optional[str] = None
    act: Optional[int] = None
    bcs: List[int] = []  # map


class Iron(BaseModel):
    """
    Class for the iron yoke data
    """

    key_points: Dict[str, Coord] = {}
    hyper_lines: Dict[str, HyperLine] = {}
    hyper_areas: Dict[str, HyperArea] = {}
    hyper_holes: Dict[int, HyperHole] = {}


class Cadata(BaseModel):
    """
    Class for the conductor data
    """

    insul: Dict[str, Insulation] = {}
    filament: Dict[str, Filament] = {}
    strand: Dict[str, Strand] = {}
    transient: Dict[str, Transient] = {}
    quench: Dict[str, Quench] = {}
    cable: Dict[str, Cable] = {}
    conductor: Dict[str, Conductor] = {}


class Coil(BaseModel):
    """
    Class for the coil data
    """

    blocks: Dict[str, Block] = {}
    groups: Dict[str, Group] = {}
    transs: Dict[str, Trans] = {}


class StrandGroup(BaseModel):
    """
    Class for strand group
    """

    strand_positions: Dict[int, Coord] = {}


class Corner(BaseModel):
    """
    Class for corner positions
    """

    iH: Coord = Coord()  # inner left
    iL: Coord = Coord()  # inner right
    oH: Coord = Coord()  # outer left
    oL: Coord = Coord()  # outer right


class HalfTurnCorner(BaseModel):
    """
    Class for corner type
    """

    insulated: Corner = Corner()
    bare: Corner = Corner()


class HalfTurn(BaseModel):
    """
    Class for half-turn data
    """

    corners: HalfTurnCorner = HalfTurnCorner()
    strand_groups: Dict[int, StrandGroup] = {}


class Order(BaseModel):
    """
    Class for electrical order (block location)
    """

    coil: Optional[int] = None
    pole: Optional[int] = None
    layer: Optional[int] = None
    winding: Optional[int] = None
    block: Optional[int] = None


class CenterShift(BaseModel):
    """
    Class for bore center shift
    """

    inner: Coord = Coord()
    outer: Coord = Coord()


class Wedge(BaseModel):
    """
    Class for wedge positions
    """

    corners: Corner = Corner()
    corners_ins: Corner = Corner()
    corrected_center: CenterShift = CenterShift()
    corrected_center_ins: CenterShift = CenterShift()
    order_l: Order = Order()
    order_h: Order = Order()


class BlockData(BaseModel):
    """
    Class for block data
    """

    block_corners: Corner = Corner()
    block_corners_ins: Corner = Corner()
    current_sign: Optional[int] = None
    half_turns: Dict[int, HalfTurn] = {}


class WindingData(BaseModel):
    """
    Class for winding data
    """

    blocks: Dict[int, BlockData] = {}
    conductor_name: Optional[str] = None
    conductors_number: Optional[int] = None


class Winding(BaseModel):
    """
    Class for windings
    """

    windings: Dict[int, WindingData] = {}


class Layer(BaseModel):
    """
    Class for winding layers
    """

    layers: Dict[int, Winding] = {}


class Pole(BaseModel):
    """
    Class for poles
    """

    poles: Dict[int, Layer] = {}
    type: Optional[str] = None
    bore_center: Coord = Coord()


class CoilData(BaseModel):
    """
    Class for coils
    """

    coils: Dict[int, Pole] = {}
    physical_order: List[Order] = []


class RoxieRawData(BaseModel):
    """
    Class for the raw data
    """

    cadata: Cadata = Cadata()
    coil: Coil = Coil()


class RoxieData(BaseModel):
    """
    Class for the roxie parser
    """

    iron: Iron = Iron()
    coil: CoilData = CoilData()
    wedges: Dict[int, Wedge] = {}



import os
import pathlib
import subprocess
import logging
import timeit
import re

import gmsh
from fiqus.pro_assemblers.ProAssembler import ASS_PRO as aP
from fiqus.utils.Utils import GmshUtils
from fiqus.utils.Utils import FilesAndFolders as Util
from fiqus.data import DataFiQuS as dF
from fiqus.data import RegionsModelFiQuS as rM
from fiqus.data import DataMultipole as dM

logger = logging.getLogger(__name__)


class AssignNaming:
    def __init__(self, data: dF.FDM() = None):
        """
        Class to assign naming convention
        :param data: FiQuS data model
        """
        self.data: dF.FDM() = data

        self.naming_conv = {'omega': 'Omega', 'boundary': 'Bd_', 'powered': '_p', 'induced': '_i', 'air': '_a',
                            'air_far_field': '_aff', 'iron': '_bh', 'conducting': '_c', 'insulator': '_ins', 'terms': 'Terms'}
        self.data.magnet.postproc.electromagnetics.volumes = \
            [self.naming_conv['omega'] + (self.naming_conv[var] if var != 'omega' else '') for var in self.data.magnet.postproc.electromagnetics.volumes]
        self.data.magnet.postproc.thermal.volumes = \
            [self.naming_conv['omega'] + (self.naming_conv[var] if var != 'omega' else '') for var in self.data.magnet.postproc.thermal.volumes]


class RunGetdpMultipole:
    def __init__(self, data: AssignNaming = None, solution_folder: str = None, GetDP_path: str = None, verbose: bool = False):
        """
        Class to solve pro file
        :param data: FiQuS data model
        :param GetDP_path: settings data model
        :param verbose: If True more information is printed in python console.
        """
        logger.info(
            f"Initializing Multipole runner for {os.path.basename(solution_folder)}."
        )
        self.data: dF.FDM() = data.data
        self.naming_conv: dict = data.naming_conv
        self.solution_folder = solution_folder
        self.GetDP_path = GetDP_path
        self.verbose: bool = verbose
        self.call_method = 'subprocess'  # or onelab

        self.rm_EM = rM.RegionsModel()
        self.rm_TH = rM.RegionsModel()
        self.rc = dM.MultipoleRegionCoordinate()\
            if self.data.magnet.mesh.thermal.isothermal_conductors and self.data.magnet.solve.thermal.solve_type else None

        self.gu = GmshUtils(self.solution_folder, self.verbose)
        self.gu.initialize(verbosity_Gmsh=self.data.run.verbosity_Gmsh)
        self.occ = gmsh.model.occ
        self.mesh = gmsh.model.mesh

        self.brep_iron_curves = {1: set(), 2: set(), 3: set(), 4: set()}
        self.mesh_files = os.path.join(os.path.dirname(self.solution_folder), self.data.general.magnet_name)
        self.model_file = os.path.join(self.solution_folder, 'Center_line.csv')

        self.mf = {'EM': f"{self.mesh_files}_EM.msh", 'TH': f"{self.mesh_files}_TH.msh"}

    def loadRegionFiles(self):
        if self.data.magnet.solve.electromagnetics.solve_type:
            self.rm_EM = Util.read_data_from_yaml(f"{self.mesh_files}_EM.reg", rM.RegionsModel)
        if self.data.magnet.solve.thermal.solve_type:
            self.rm_TH = Util.read_data_from_yaml(f"{self.mesh_files}_TH.reg", rM.RegionsModel)

    def loadRegionCoordinateFile(self):
        self.rc = Util.read_data_from_yaml(f"{self.mesh_files}_TH.reco", dM.MultipoleRegionCoordinate)

    def assemblePro(self):
        logger.info(f"Assembling pro file...")
        start_time = timeit.default_timer()
        ap = aP(file_base_path=os.path.join(self.solution_folder, self.data.general.magnet_name), naming_conv=self.naming_conv)
        BH_curves_path = os.path.join(pathlib.Path(os.path.dirname(__file__)).parent, 'pro_material_functions', 'ironBHcurves.pro')
        ap.assemble_combined_pro(template='Multipole_template.pro', rm_EM=self.rm_EM, rm_TH=self.rm_TH, rc=self.rc, dm=self.data, mf=self.mf, BH_curves_path=BH_curves_path)
        logger.info(
            f"Assembling pro file took"
            f" {timeit.default_timer() - start_time:.2f} s."
        )

    def solve_and_postprocess(self):
        commands = None
        if self.call_method == 'onelab':
            commands = f"-solve -v2 -pos -verbose {self.data.run.verbosity_GetDP}"
        elif self.call_method == 'subprocess':
            commands = []
            commands.append(["-solve", 'resolution', "-v2", "-pos", "Dummy", "-verbose", str(self.data.run.verbosity_GetDP)])

        self._run(commands=commands)

    def postprocess(self):
        if self.call_method == 'onelab':
            command = "-v2 -pos -verbose {self.data.run.verbosity_GetDP}"
        elif self.call_method == 'subprocess':
            command = [["-v2", "-pos", "-verbose", str(self.data.run.verbosity_GetDP)]]
        self._run(commands=command)

    def _run(self, commands):
        logger.info("Solving...")
        start_time = timeit.default_timer()
        if self.call_method == 'onelab':
            for command in commands:
                gmsh.onelab.run(f"{self.data.general.magnet_name}",
                                f"{self.GetDP_path} {os.path.join(self.solution_folder, self.data.general.magnet_name)}.pro {command}")
            gmsh.onelab.setChanged("GetDP", 0)
        elif self.call_method == 'subprocess':
            # subprocess.call([f"{self.GetDP_path}", f"{os.path.join(self.solution_folder, self.data.general.magnet_name)}.pro"] + command + ["-msh", f"{self.mesh_files}.msh"])

            # view_tag = gmsh.view.getTags()  # this should be b
            # # # v = "View[" + str(gmsh.view.getIndex('b')) + "]"
            # gmsh.view.write(view_tag, f"{os.path.join(self.solution_folder, self.data.general.magnet_name)}-view.msh")

            if self.data.magnet.solve.noOfMPITasks:
                mpi_prefix = ["mpiexec", "-np", str(self.data.magnet.solve.noOfMPITasks)]
            else:
                mpi_prefix = []
            
            for command in commands:
                getdpProcess = subprocess.Popen(
                    mpi_prefix + [f"{self.GetDP_path}", f"{os.path.join(self.solution_folder, self.data.general.magnet_name)}.pro"] +
                    command,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                )
                with getdpProcess.stdout:
                    for line in iter(getdpProcess.stdout.readline, b""):
                        line = line.decode("utf-8").rstrip()
                        line = line.split("\r")[-1]
                        if "Info" in line:
                            parsedLine = re.sub(r"Info\s*:\s*", "", line)
                            logger.info(parsedLine)
                        elif "Warning" in line:
                            parsedLine = re.sub(r"Warning\s*:\s*", "", line)
                            if "Unknown" not in parsedLine:
                                logger.warning(parsedLine)
                        elif "Error" in line:
                            parsedLine = re.sub(r"Error\s*:\s*", "", line)
                            logger.error(parsedLine)
                            raise Exception(parsedLine)
                        elif "Critical" in line:
                            parsedLine = re.sub(r"Critical\s*:\s*", "", line)
                            logger.critical(parsedLine)
                        # catch the maximum temperature line
                        elif "Maximum temperature" in line:
                            parsedLine = re.sub(r"Print\s*:\s*", "", line)
                            logger.info(parsedLine)
                        # this activates the debugging message mode
                        elif self.data.run.verbosity_GetDP > 99:
                            logger.info(line)

                getdpProcess.wait()

        logger.info(
            f"Solving took {timeit.default_timer() - start_time:.2f} s."
        )

    def ending_step(self, gui: bool = False):
        if gui:
            self.gu.launch_interactive_GUI()
        else:
            gmsh.clear()
            gmsh.finalize()



from pydantic import BaseModel, Field
from typing import List, Literal, Optional


# ============= GEOMETRY ============= #
# -- Input/Output settings -- #
class CACStrandIOsettingsLoad(BaseModel):
    """
    Level 3: Class for Input/Output settings for the cable geometry
    """

    load_from_yaml: Optional[bool] = Field(
        default=False,
        description="True to load the geometry from a YAML file, false to generate the geometry.",
    )
    filename: Optional[str] = Field(
        default=None,
        description="Name of the YAML file from which to load the geometry.",
    )

class CACStrandIOsettingsSave(BaseModel):
    """
    Level 3: Class for Input/Output settings for the cable geometry
    """

    save_to_yaml: Optional[bool] = Field(
        default=False,
        description="True to save the geometry to a YAML-file, false to not save the geometry.",
    )
    filename: Optional[str] = Field(
        default=None,
        description="Name of the output geometry YAML file.",
    )

class CACStrandGeometryIOsettings(BaseModel):
    """
    Level 2: Class for Input/Output settings for the cable geometry
    """

    load: CACStrandIOsettingsLoad = (
        CACStrandIOsettingsLoad()
    )
    save: CACStrandIOsettingsSave = (
        CACStrandIOsettingsSave()
    )

# -- Strand geometry parameters -- #
class CACStrandGeometry(BaseModel):
    """
    Level 2: Class for strand geometry parameters
    """
    io_settings: CACStrandGeometryIOsettings = CACStrandGeometryIOsettings()
    hexagonal_filaments: Optional[bool] = Field(
        default=None,
        description="Field for specifying the shape of the filaments. True for hexagonal, False for circular.",
    )
    filament_circular_distribution: Optional[bool] = Field(
        default=None,
        description="Field for specifying the geometrical distribution of the filaments. Set True to distribute the filaments in a circular pattern and False to distribute them in a hexagonal pattern."
    )
    air_radius: Optional[float] = Field(
        default=None, description="Radius of the circular numerical air region (m)."
    )


# ============= MESH ============= #
# -- Filament mesh settings -- #
class CACStrandMeshFilaments(BaseModel):
    """
    Level 3: Class for FiQuS ConductorAC
    """

    boundary_mesh_size_ratio: Optional[float] = Field(
        default=None,
        description="Mesh size at filament boundaries, relative to the radius of the filaments. E.g. 0.1 means that the mesh size is 0.1 times the filament radius.",
    )
    center_mesh_size_ratio: Optional[float] = Field(
        default=None,
        description="Mesh size at filament center, relative to the radius of the filaments. E.g. 0.1 means that the mesh size is 0.1 times the filament radius.",
    )
    amplitude_dependent_scaling: Optional[bool] = Field(
        default=False,
        description="Amplitude dependent scaling uses the field amplitude to approximate the field penetration distance in the filaments to alter the filament mesh. If the field penetration distance is low (i.e. for low field amplitudes) this feature increases mesh density in the region where the field is expected to penetrate, and decreases the mesh resolution in the region where the field does not penetrate.",
    )
    field_penetration_depth_scaling_factor: Optional[float] = Field(
        default=None,
        description="Scaling factor for the estimate of the field penetration depth, used for amplitude dependent scaling. ",
    )
    desired_elements_in_field_penetration_region: Optional[float] = Field(
        default=None,
        description="Desired number of elements in the field penetration region. This parameter is used for amplitude dependent scaling, and determines the number of elements in the region where the field is expected to penetrate.",
    )

# -- Matrix mesh settings -- #
class CACStrandMeshMatrix(BaseModel):
    """
    Level 3: Class for FiQuS ConductorAC
    """

    mesh_size_matrix_ratio_inner: Optional[float] = Field(
        default=None,
        description="Mesh size at the matrix center, relative to the filament radius.",
    )
    mesh_size_matrix_ratio_middle: Optional[float] = Field(
        default=None,
        description="Mesh size at the matrix middle partition, relative to the filament radius.",
    )
    mesh_size_matrix_ratio_outer: Optional[float] = Field(
        default=None,
        description="Mesh size at the matrix outer boundary, relative to the filament radius.",
    )
    interpolation_distance_from_filaments_ratio: Optional[float] = Field(
        default=None,
        description="The mesh size is interpolated from the filament boundaries into the matrix, over a given distance. This parameter determines the distance over which the mesh size is interpolated.",
    )
    rate_dependent_scaling_matrix: Optional[bool] = Field(
        default=False,
        description="Rate dependent scaling uses the expected skin depth in the matrix to determine the matrix mesh. If the skin depth is low (i.e. for high frequencies) this feature increases mesh density in the region where the current is expected to flow, while decreasing the mesh resolution in the region where the current is not expected to flow.",
    )
    skindepth_scaling_factor: Optional[float] = Field(
        default=None,
        description="Scaling factor for the estimate of the skin depth, used for rate dependent scaling.",
    )
    desired_elements_in_skindepth: Optional[float] = Field(
        default=None, description="Desired number of elements in the skin depth region. This parameter is used for rate dependent scaling, and determines the number of elements in the region where the current is expected to flow."
    )

# -- Air mesh settings -- #
class CACStrandMeshAir(BaseModel):
    """
    Level 3: Class for FiQuS ConductorAC
    """

    max_mesh_size_ratio: Optional[float] = Field(
        default=None,
        description="Mesh size at the outer boundary of the air region, relative to the filament radius. E.g. 10 means that the mesh size is 10 times the filament radius.",
    )

# -- Strand mesh settings -- #
class CACStrandMesh(BaseModel):
    """
    Level 2: Class for FiQuS ConductorAC
    """

    scaling_global: Optional[float] = Field(
        default=1, description="Global scaling factor for mesh size."
    )
    filaments: CACStrandMeshFilaments = CACStrandMeshFilaments()
    matrix: CACStrandMeshMatrix = CACStrandMeshMatrix()
    air: CACStrandMeshAir = CACStrandMeshAir()


# ============= SOLVE ============= #
# -- General parameters -- #
class CACStrandSolveGeneralparameters(BaseModel):
    """
    Level 3: Class for general parameters
    """
    temperature: float = Field(default=1.9, description="Temperature (K) of the strand.")
    superconductor_linear: Optional[bool] = Field(default=False, description="For debugging: replace LTS by normal conductor.")


# -- Initial conditions -- #
class CACStrandSolveInitialconditions(BaseModel):
    """
    Level 3: Class for initial conditions
    """

    init_from_pos_file: bool = Field(
        default=False, description="This field is used to initialize the solution from a non-zero field solution stored in a .pos file."
    )
    pos_file_to_init_from: Optional[str] = Field(
        default=None,
        description="Name of .pos file for magnetic field (A/m) from which the solution should be initialized."
        " Should be in the Geometry_xxx/Mesh_xxx/ folder in which the Solution_xxx will be saved.",
    )


# -- Source parameters -- #
class CACStrandSolveSourceparametersSineSuperimposedDC(BaseModel):
    """
    Level 5: Class for superimposed DC field or current parameters for the sine source
    """
    field_magnitude: Optional[float] = Field(default=0.0, description="DC field magnitude (T) (direction along y-axis). Solution must be initialized with a non-zero field solution stored in a .pos file if non-zero DC field is used.")
    current_magnitude: Optional[float] = Field(default=0.0, description="DC current magnitude (A). Solution must be initialized with a non-zero field solution stored in a .pos file if non-zero DC current is used.")

class CACStrandSolveSourceparametersSine(BaseModel):
    """
    Level 4: Class for Sine source parameters
    """
    frequency: Optional[float] = Field(default=None, description="Frequency of the sine source (Hz).")
    field_amplitude: Optional[float] = Field(default=None, description="Amplitude of the sine field (T).")
    current_amplitude: Optional[float] = Field(default=None, description="Amplitude of the sine current (A).")
    field_angle: Optional[float] = Field(default=90, description="Angle of the sine field direction, with respect to the x-axis (degrees).")
    superimposed_DC: CACStrandSolveSourceparametersSineSuperimposedDC = CACStrandSolveSourceparametersSineSuperimposedDC()

class CACStrandSolveSourceparametersPiecewise(BaseModel):
    """
    Level 4: Class for piecewise (linear) source parameters
    """
    source_csv_file: Optional[str] = Field(default=None, description="File name for the from_file source type defining the time evolution of current and field (in-phase). Multipliers are used for each of them. The file should contain two columns: 'time' (s) and 'value' (field/current (T/A)), with these headers. If this field is set, times, applied_fields_relative and transport_currents_relative are ignored.")
    times: Optional[List[float]] = Field(default=None, description="Time instants (s) defining the piecewise linear sources. Used only if source_csv_file is not set. Can be scaled by time_multiplier.")
    applied_fields_relative: Optional[List[float]] = Field(default=None, description="Applied fields relative to multiplier applied_field_multiplier at the time instants 'times'. Used only if source_csv_file is not set.")
    transport_currents_relative: Optional[List[float]] = Field(default=None, description="Transport currents relative to multiplier transport_current_multiplier at the time instants 'times'. Used only if source_csv_file is not set.")
    time_multiplier: Optional[float] = Field(default=None, description="Multiplier for the time values in times (scales the time values). Also used for the time values in the source_csv_file.")
    applied_field_multiplier: Optional[float] = Field(default=None, description="Multiplier for the applied fields in applied_fields_relative. Also used for the values in the source_csv_file.")
    transport_current_multiplier: Optional[float] = Field(default=None, description="Multiplier for the transport currents in transport_currents_relative. Also used for the values in the source_csv_file.")

class CACStrandSolveSourceparameters(BaseModel):
    """
    Level 3: Class for material properties
    """

    source_type: Literal['sine', 'piecewise'] = Field(
        default='sine',
        description="Time evolution of applied current and magnetic field. Supported options are: sine, sine_with_DC, piecewise_linear, from_list.",
    )
    boundary_condition_type: str = Field(
        default="Natural",
        description="Boundary condition type. Supported options are: Natural, Essential. Do not use essential boundary condition with induced currents.",
    )
    sine: CACStrandSolveSourceparametersSine = CACStrandSolveSourceparametersSine()
    piecewise: CACStrandSolveSourceparametersPiecewise = CACStrandSolveSourceparametersPiecewise()


# -- Numerical parameters -- #
class CACStrandSolveNumericalparametersSine(BaseModel):
    """ 
    Level 4: Numerical parameters corresponding to the sine source
    """
    timesteps_per_period: Optional[float] = Field(default=None, description="Initial value for number of time steps (-) per period for the sine source. Determines the initial time step size.")
    number_of_periods_to_simulate: Optional[float] = Field(default=None, description="Number of periods (-) to simulate for the sine source.")

class CACStrandSolveNumericalparametersPiecewise(BaseModel):
    """
    Level 4: Numerical parameters corresponding to the piecewise source
    """
    time_to_simulate: Optional[float] = Field(default=None, description="Total time to simulate (s). Used for the piecewise source.")
    timesteps_per_time_to_simulate: Optional[float] = Field(default=None, description="If variable_max_timestep is False. Number of time steps (-) per period for the piecewise source.")
    force_stepping_at_times_piecewise_linear: bool = Field(default=False, description="If True, time-stepping will contain exactly the time instants that are in the times_source_piecewise_linear list (to avoid truncation maximum applied field/current values).")

    variable_max_timestep: bool = Field(default=False, description="If False, the maximum time step is kept constant through the simulation. If True, it varies according to the piecewise definition.")
    times_max_timestep_piecewise_linear: Optional[List[float]] = Field(default=None, description="Time instants (s) defining the piecewise linear maximum time step.")
    max_timestep_piecewise_linear: Optional[List[float]] = Field(default=None, description="Maximum time steps (s) at the times_max_timestep_piecewise_linear. Above the limits, linear extrapolation of the last two values.")

class CACStrandSolveNumericalparameters(BaseModel):
    """
    Level 3: Class for numerical parameters
    """

    sine: CACStrandSolveNumericalparametersSine = CACStrandSolveNumericalparametersSine()
    piecewise: CACStrandSolveNumericalparametersPiecewise = CACStrandSolveNumericalparametersPiecewise()


# -- Formulation parameters -- #
class CACStrandSolveFormulationparameters(BaseModel):
    """
    Level 3: Class for finite element formulation parameters
    """

    formulation: Literal['voltage_based'] = Field(
        default='voltage_based',
        description="Currently, only possible option is voltage_based."
    )
    dynamic_correction: Optional[bool] = Field(
        default=True,
        description="In the voltage_based case, do we activate the dynamic correction?",
    )
    compute_temperature: Optional[bool] = Field(
        default=False, description="Do we compute the temperature?"
    )

    two_ell_periodicity : Optional[bool] = Field(
        default=True, description="True to integrate over twice the shortest periodicity length, False to integrate over the shortest periodicity length. "
    )


class CACStrandSolve(BaseModel):
    """
    Level 2: Class for FiQuS ConductorAC Strand solver settings
    """
    pro_template: Optional[Literal['ConductorAC_template.pro']] = Field(
        default='ConductorAC_template.pro',
        description="Name of the .pro template file."
    )
    conductor_name: Optional[str] = Field(
        default=None, description="Name of the conductor. Must match a conductor name in the conductors section of the input YAML-file."
    )
    formulation_parameters: CACStrandSolveFormulationparameters = (
        CACStrandSolveFormulationparameters()
    )
    general_parameters: CACStrandSolveGeneralparameters = (
        CACStrandSolveGeneralparameters()
    )
    # material_properties: CACStrandSolveMaterialproperties = (
    #     CACStrandSolveMaterialproperties()
    # )
    initial_conditions: CACStrandSolveInitialconditions = (
        CACStrandSolveInitialconditions()
    )
    source_parameters: CACStrandSolveSourceparameters = (
        CACStrandSolveSourceparameters()
    )
    numerical_parameters: CACStrandSolveNumericalparameters = (
        CACStrandSolveNumericalparameters()
    )


# ============= POSTPROC ============= #
class CACStrandPostprocBatchpostprocLossMapCrossSection(BaseModel):
    """
    Level 5: Class with settings for plotting a cross-section of the loss map.
    """
    plot_cross_section: bool = Field(
        default=False, description="Set True to plot a cross-section of the loss map."
    )
    save_plot: bool = Field(default=False, description="Set True to save the plot.")
    filename: str = Field(default="cross_section", description="Name of the plot file.")
    axis_to_cut: str = Field(
        default="x", description="Axis to cut for the cross-section."
    )
    cut_value: float = Field(
        default=0, description="Value of the axis to cut for the cross-section."
    )

    ylabel: str = Field(default="Loss", description="Label of the y-axis.")
    title: Optional[str] = Field(
        default=None,
        description="Title of the plot. The placeholder <<cut_value>> can be used to indicate the value of the cut axis.",
    )


class CACStrandPostprocBatchpostprocLossMapCrossSectionSweep(BaseModel):
    """
    Level 5: Class with settings for animating a cross-section sweep of the loss map along one axis.
    """
    animate_cross_section_sweep: bool = Field(
        default=False,
        description="Set True to animate a cross-section sweep of the loss map along one axis.",
    )
    save_plot: bool = Field(
        default=False, description="Set True to save the animation."
    )
    filename: str = Field(
        default="crossSectionSweep", description="Name of the animation file."
    )
    axis_to_sweep: str = Field(
        default="x", description="Axis to sweep for the animation."
    )
    ylabel: str = Field(default="Loss", description="Label of the y-axis.")
    title: Optional[str] = Field(
        default=None,
        description="Title of the plot. Use <<sweep_value>> to indicate the value of the sweep axis.",
    )


class CACStrandPostprocBatchpostprocLossMap(BaseModel):
    """
    Level 4: Class with settings for generating loss maps
    """
    produce_loss_map: bool = Field(
        default=False, description="Set True to produce a loss map."
    )
    save_plot: bool = Field(default=False, description="Set True to save the plot.")
    filename: str = Field(default="loss_map", description="Name of the plot file.")
    x_val: Optional[str] = Field(
        default=None, description="Parameter to be plotted on the x-axis. This field corresponds to a parameter in the input YAML-file. E.g. 'solve.source_parameters.sine.frequency' will plot the loss map for different frequencies."
    )
    y_val: Optional[str] = Field(
        default=None, description="Parameter to be plotted on the y-axis. This field corresponds to a parameter in the input YAML-file. E.g. 'solve.source_parameters.sine.field_amplitude' will plot the loss map for different applied field amplitudes."
    )
    x_steps: int = Field(default=20, description="Number of steps on the x-axis.")
    y_steps: int = Field(default=20, description="Number of steps on the y-axis.")
    loss_type: Literal['TotalLoss', 'FilamentLoss', 'CouplingLoss', 'EddyLoss'] = Field(
        default='TotalLoss',
        description="Type of loss to be plotted. Supported options are: TotalLoss, FilamentLoss, CouplingLoss, EddyLoss."
    )
    x_log: bool = Field(
        default=True, description="Set True to plot x-axis in log-scale."
    )
    y_log: bool = Field(
        default=True, description="Set True to plot y-axis in log-scale."
    )
    loss_log: bool = Field(
        default=True, description="Set True to plot loss in log-scale."
    )
    x_norm: float = Field(default=1, description="Normalization factor for x-axis.")
    y_norm: float = Field(default=1, description="Normalization factor for y-axis.")
    loss_norm: float = Field(default=1, description="Normalization factor for the AC-loss.")
    show_datapoints: bool = Field(
        default=True, description="Set True to show markers for all the datapoints in the loss map."
    )

    title: Optional[str] = Field(default=None, description="Title for the plot.")
    xlabel: Optional[str] = Field(default=None, description="Label for the x-axis.")
    ylabel: Optional[str] = Field(default=None, description="Label for the y-axis.")

    # lossType_dominance_contour: CACStrandPostprocBatchpostprocLossMapDominanceCountour = (
    #     CACStrandPostprocBatchpostprocLossMapDominanceCountour()
    # )

    show_loss_type_dominance_contour: bool = Field(
        default=False,
        description="Set True to plot a contour curve separating regions where different loss types dominate. ",
    )

    cross_section: CACStrandPostprocBatchpostprocLossMapCrossSection = (
        CACStrandPostprocBatchpostprocLossMapCrossSection()
    )
    cross_section_sweep: CACStrandPostprocBatchpostprocLossMapCrossSectionSweep = (
        CACStrandPostprocBatchpostprocLossMapCrossSectionSweep()
    )


class CACStrandPostprocBatchpostprocPlot2d(BaseModel):
    """
    Level 4: Class for 2D plot settings
    """
    produce_plot2d: bool = Field(
        default=False, description="Set True to produce a 2D plot."
    )
    combined_plot: bool = Field(
        default=False,
        description="Set True to produce a combined plot for all simulations. If False, a separate plot is produced for each simulation.",
    )
    save_plot: bool = Field(default=False, description="Set True to save the plot.")
    filename: str = Field(default="plot2d", description="Name of the plot file.")
    x_val: Optional[str] = Field(
        default=None, description="Value to be plotted on the x-axis. Parameters in the input YAML-file and class-variables from the plotter 'SimulationData' class can be accessed trough the notation << . >>. E.g. '<<solve.source_parameters.sine.frequency>>' will create a 2D plot with frequency on the x-axis. '<<time>>' will create a plot with time on the x-axis."
    )
    y_vals: Optional[List[str]] = Field(
        default=None, description=" List of values to be plotted on the y-axis. Parameters in the input YAML-file and class-variables from the plotter 'SimulationData' class can be accessed trough the notation << . >>. E.g. total AC-loss per cycle can be accessed as ['<<total_power_per_cycle['TotalLoss_dyn']>>']."
    )
    labels: Optional[List[str]] = Field(
        default=None,
        description="List of labels for the plot. Each label corresponding to a value in y_val.",
    )
    linestyle: Optional[str] = Field(
        default=None, description="Linestyle for the plot."
    )

    title: Optional[str] = Field(default=None, description="Title for the plot.")
    xlabel: Optional[str] = Field(default=None, description="Label for the x-axis.")
    ylabel: Optional[str] = Field(default=None, description="Label for the y-axis.")
    x_log: bool = Field(default=False, description="Set True to plot x-axis in log-scale.")
    y_log: bool = Field(default=False, description="Set True to plot y-axis in log-scale.")
    legend: bool = Field(default=True, description="Set True to show legend.")


class CACStrandPostprocBatchpostprocFilter(BaseModel):
    """
    Level 4: Field for filtering simulations based on simulation parameters for batch post-processing
    """
    apply_filter: bool = Field(
        default=False,
        description="Set True to filter simulations by parameters from the input YAML-file.",
    )
    filter_criterion: Optional[str] = Field(
        default=None,
        description="Criterion used to filter simulations based on simulation parameters. For example will '<<solve.source_parameters.sine.frequency>> > 100' disregard simulations done with frequencies lower than 100Hz.",
    )


class CACStrandPostprocBatchpostprocSort(BaseModel):
    """
    Level 4: Field for sorting simulations based on simulation parameters for batch post-processing
    """
    apply_sort: bool = Field(default=False, description="Set True to sort simulations.")
    sort_key: Optional[str] = Field(
        default=None,
        description="Criterion used to sort simulations based on simulation parameters. For example will 'sd.total_power_per_cycle['TotalLoss'] sort simulations based on the total loss.",
    )


class CACStrandPostprocBatchpostproc(BaseModel):
    """
    Level 3: Class for batch post-processing settings
    """
    postProc_csv: Optional[str] = Field(
        default=None,
        description="Name of the .csv file for post-processing (without file extension). This file specifies the simulations to be post-processed. The file is structured into three columns, specifying the folder names to access the simulation results: 'input.run.geometry', 'input.run.mesh' and 'input.run.solve'. Each row corresponds to a simulation to be post-processed.",
    )
    output_folder: Optional[str] = Field(
        default=None,
        description="Batch post-processing creates a folder with the given name in the output directory, where all the plots are saved.",
    )
    filter: CACStrandPostprocBatchpostprocFilter = CACStrandPostprocBatchpostprocFilter()
    sort: CACStrandPostprocBatchpostprocSort = CACStrandPostprocBatchpostprocSort()
    loss_map: CACStrandPostprocBatchpostprocLossMap = CACStrandPostprocBatchpostprocLossMap()
    plot2d: CACStrandPostprocBatchpostprocPlot2d = CACStrandPostprocBatchpostprocPlot2d()


class CACStrandPostprocPlotInstPower(BaseModel):
    """
    Level 3: Class with settings for generating plots of instantaneous power
    """
    show: bool = Field(default=False, description="Creates a plot for the calculated instantaneous AC loss (W/m) as a function of time (s).")
    title: str = Field(default="Instantaneous Power", description="Title for the plot.")
    save: bool = Field(default=False, description="Set True to save the plot.")
    save_file_name: str = Field(
        default="instantaneous_power", description="Name of the plot file."
    )


class CACStrandPostprocCleanup(BaseModel):
    """
    Level 3: Class for cleanup settings
    """
    remove_pre_file: bool = Field(
        default=False,
        description="Set True to remove the .pre-file after post-processing, to save disk space.",
    )
    remove_res_file: bool = Field(
        default=False,
        description="Set True to remove the .res-file after post-processing, to save disk space.",
    )
    remove_msh_file: bool = Field(
        default=False,
        description="Set True to remove the .msh-file after post-processing, to save disk space.",
    )


class CACStrandPostproc(BaseModel):
    """
    Level 2: Class for FiQuS ConductorAC
    """

    generate_pos_files: bool = Field(
        default=True,
        description="Set True to generate .pos-files during post-processing",
    )
    plot_instantaneous_power: CACStrandPostprocPlotInstPower = (
        CACStrandPostprocPlotInstPower()
    )
    compute_current_per_filament: bool = Field(
        default=False,
        description="Computes current in every filament, with decomposition into magnetization and transport current.",
    )
    save_last_current_density: Optional[str] = Field(
        default=None,
        description="Saves the last current density field solution (out-of-plane) in the file given as a string."
        " The '.pos' extension will be appended to it. Nothing is done if None."
        " This can be for using the current density as an initial condition (but not implemented yet).",
    )
    save_last_magnetic_field: Optional[str] = Field(
        default=None,
        description="Saves the last magnetic field solution (in-plane) in the file given as a string."
        " The '.pos' extension will be appended to it. Nothing is done if None."
        " This is for using the magnetic field as an initial condition for another resolution.",
    )
    cleanup: CACStrandPostprocCleanup = CACStrandPostprocCleanup()
    batch_postproc: CACStrandPostprocBatchpostproc = CACStrandPostprocBatchpostproc()

# ============= BASE ============= #
class CACStrand(BaseModel):
    """
    Level 1: Class for FiQuS ConductorAC
    """

    type: Literal["CACStrand"]
    geometry: CACStrandGeometry = CACStrandGeometry()
    mesh: CACStrandMesh = CACStrandMesh()
    solve: CACStrandSolve = CACStrandSolve()
    postproc: CACStrandPostproc = CACStrandPostproc()
    
